
from __future__ import print_function

import unittest
import weakref
import gc

from ..client.raw import Context, Cancelled
from ..wrapper import Value, Type
from .utils import RefTestCase


class TestRequest(RefTestCase):

    def testEmpty(self):
        self.assertListEqual(Context.makeRequest("").tolist(), [])

    def testValue(self):
        self.assertListEqual(Context.makeRequest("field(value)").tolist(),
                             [('field', [('value', [])])]
                             )

    def testAll(self):
        self.assertListEqual(Context.makeRequest("field()").tolist(),
                             [('field', [])]
                             )


class TestProviders(RefTestCase):

    def tearDown(self):
        gc.collect()  # try to provoke any crashes here so they can be associated with this testcase

    def testProviders(self):
        providers = Context.providers()
        self.assertIn('pva', providers)


class TestPVA(RefTestCase):

    def setUp(self):
        super(TestPVA, self).setUp()
        self.ctxt = Context("pva")

    def tearDown(self):
        self.ctxt.close()
        self.ctxt = None
        gc.collect()
        super(TestPVA, self).tearDown()

    def testGetAbort(self):
        _X = [None]

        def fn(V):
            _X[0] = V
        op = self.ctxt.get("completelyInvalidChannelName", fn)

        op.close()

        self.assertIsInstance(_X[0], Cancelled)

    def testGetAbortGC(self):
        _X = [None]

        def fn(V):
            _X[0] = V
        op = self.ctxt.get("completelyInvalidChannelName", fn)

        W = weakref.ref(op)
        del op
        gc.collect()

        self.assertIsNone(W())

        self.assertIsNone(_X[0])

    def testGCCycle(self):
        _X = [None]

        def fn(V):
            _X[0] = V
        op = self.ctxt.get("completelyInvalidChannelName", fn)

        fn._cycle = op  # create cycle: op -> fn -> fn.__dict__ -> op

        self.assertIn(fn.__dict__, gc.get_referrers(op))

        W = weakref.ref(op), weakref.ref(fn)
        del op, fn
        gc.collect()

        self.assertIsNone(W[0]())
        self.assertIsNone(W[1]())

        self.assertIsNone(_X[0])

    def testRPCAbort(self):
        P = Value(Type([
            ('value', 'i'),
        ]), {
            'value': 42,
        })

        _X = [None]

        def fn(V):
            _X[0] = V
        op = self.ctxt.rpc("completelyInvalidChannelName", fn, P)

        W = weakref.ref(op)
        del op
        gc.collect()

        self.assertIsNone(W())

        self.assertIsNone(_X[0])

    def testMonAbort(self):
        canery = object()
        _X = [canery]

        def evt(V):
            _X[0] = V

        op = self.ctxt.monitor("completelyInvalidChannelName", evt)

        op.close()

        self.assertIsInstance(_X[0], Cancelled)

    def testMonCycle(self):
        canery = object()
        _X = [canery]

        def evt(V):
            _X[0] = V

        op = self.ctxt.monitor("completelyInvalidChannelName", evt)

        evt._cycle = op  # op -> evt -> evt.__dict__ -> op

        self.assertIn(evt.__dict__, gc.get_referrers(op))

        W = weakref.ref(op)
        del op, evt
        gc.collect()

        self.assertIsNone(W())

        self.assertIs(_X[0], canery)




import time
import sys
import numpy

from ..wrapper import Type, Value
from .common import alarm, timeStamp

if sys.version_info >= (3, 0):
    unicode = str


class ntwrappercommon(object):
    raw = timestamp = None

    def _store(self, value):
        assert isinstance(value, Value), value
        self.raw = value
        self.severity = value.get('alarm.severity', 0)
        self.status = value.get('alarm.status', 0)
        S, NS = value.get('timeStamp.secondsPastEpoch', 0), value.get('timeStamp.nanoseconds', 0)
        self.raw_stamp = S, NS
        self.timestamp = S + NS * 1e-9
        # TODO: unpack display/control
        return self

    def __str__(self):
        return '%s %s' % (time.ctime(self.timestamp), self.__repr__())

    tostr = __str__

class ntfloat(ntwrappercommon, float):
    """
    Augmented float with additional attributes

    * .severity
    * .status
    * .timestamp - Seconds since 1 Jan 1970 UTC as a float
    * .raw_stamp - A tuple (seconds, nanoseconds)
    * .raw - The underlying :py:class:`p4p.Value`.
    """


class ntint(ntwrappercommon, int):
    """
    Augmented integer with additional attributes

    * .severity
    * .status
    * .timestamp - Seconds since 1 Jan 1970 UTC as a float
    * .raw_stamp - A tuple (seconds, nanoseconds)
    * .raw - The underlying :py:class:`p4p.Value`.
    """


class ntbool(ntwrappercommon, int):
    """
    Augmented boolean with additional attributes

    * .severity
    * .status
    * .timestamp - Seconds since 1 Jan 1970 UTC as a float
    * .raw_stamp - A tuple (seconds, nanoseconds)
    * .raw - The underlying :py:class:`p4p.Value`.
    """
    def __new__(cls, value):
        return int.__new__(cls, bool(value))

    def __repr__(self):
        return bool(self).__repr__().lower()


class ntstr(ntwrappercommon, unicode):
    """
    Augmented string with additional attributes

    * .severity
    * .status
    * .timestamp - Seconds since 1 Jan 1970 UTC as a float
    * .raw_stamp - A tuple (seconds, nanoseconds)
    * .raw - The underlying :py:class:`p4p.Value`.
    """


class ntnumericarray(ntwrappercommon, numpy.ndarray):
    """
    Augmented numpy.ndarray with additional attributes

    * .severity
    * .status
    * .timestamp - Seconds since 1 Jan 1970 UTC as a float
    * .raw_stamp - A tuple (seconds, nanoseconds)
    * .raw - The underlying :py:class:`p4p.Value`.
    """

    @classmethod
    def build(klass, val):
        assert len(val.shape) == 1, val.shape
        # clone
        return klass(shape=val.shape, dtype=val.dtype, buffer=val.data,
                     strides=val.strides)


class ntstringarray(ntwrappercommon, list):
    """
    Augmented list of strings with additional attributes

    * .severity
    * .status
    * .timestamp - Seconds since 1 Jan 1970 UTC as a float
    * .raw_stamp - A tuple (seconds, nanoseconds)
    * .raw - The underlying :py:class:`p4p.Value`.
    """

def _metaHelper(F, valtype, display=False, control=False, valueAlarm=False):
    isnumeric = valtype[-1:] not in '?su'
    if display and isnumeric:
        F.extend([
            ('display', ('S', None, [
                ('limitLow', valtype[-1:]),
                ('limitHigh', valtype[-1:]),
                ('description', 's'),
                ('format', 's'),
                ('units', 's'),
            ])),
        ])
    if control and isnumeric:
        F.extend([
            ('control', ('S', None, [
                ('limitLow', valtype[-1:]),
                ('limitHigh', valtype[-1:]),
                ('minStep', valtype[-1:]),
            ])),
        ])
    if valueAlarm and isnumeric:
        F.extend([
            ('valueAlarm', ('S', None, [
                ('active', '?'),
                ('lowAlarmLimit', valtype[-1:]),
                ('lowWarningLimit', valtype[-1:]),
                ('highWarningLimit', valtype[-1:]),
                ('highAlarmLimit', valtype[-1:]),
                ('lowAlarmSeverity', 'i'),
                ('lowWarningSeverity', 'i'),
                ('highWarningSeverity', 'i'),
                ('highAlarmSeverity', 'i'),
                ('hysteresis', 'd'),
            ])),
        ])

class NTScalar(object):

    """Describes a single scalar or array of scalar values and associated meta-data

    >>> stype = NTScalar('d') # scalar double
    >>> V = stype.wrap(4.2)
    >>> assert isinstance(V, Value)

    >>> stype = NTScalar.buildType('ad') # vector double
    >>> V = Value(stype, {'value': [4.2, 4.3]})

    The result of `wrap()` is an augmented value object combining
    `ntwrappercommon` and a python value type (`str`, `int`, `float`, `numpy.ndarray`).

    Agumented values have some additional attributes including:

    * .timestamp - The update timestamp is a float representing seconds since 1 jan 1970 UTC.
    * .raw_stamp - A tuple of (seconds, nanoseconds)
    * .severity - An integer in the range [0, 3]
    * .raw - The complete underlying :class:`~p4p.Value`
    """
    Value = Value

    @staticmethod
    def buildType(valtype, extra=[], display=False, control=False, valueAlarm=False):
        """Build a Type

        :param str valtype: A type code to be used with the 'value' field.  See :ref:`valuecodes`
        :param list extra: A list of tuples describing additional non-standard fields
        :param bool display: Include optional fields for display meta-data
        :param bool control: Include optional fields for control meta-data
        :param bool valueAlarm: Include optional fields for alarm level meta-data
        :returns: A :py:class:`Type`
        """
        isarray = valtype[:1] == 'a'
        F = [
            ('value', valtype),
            ('alarm', alarm),
            ('timeStamp', timeStamp),
        ]
        _metaHelper(F, valtype, display=display, control=control, valueAlarm=valueAlarm)
        F.extend(extra)
        return Type(id="epics:nt/NTScalarArray:1.0" if isarray else "epics:nt/NTScalar:1.0",
                    spec=F)

    def __init__(self, valtype='d', **kws):
        self.type = self.buildType(valtype, **kws)

    def wrap(self, value, timestamp=None):
        """Pack python value into Value

        Accepts dict to explicitly initialize fields by name.
        Any other type is assigned to the 'value' field.
        """
        if isinstance(value, Value):
            return value
        elif isinstance(value, ntwrappercommon):
            return value.raw
        elif isinstance(value, dict):
            return self.Value(self.type, value)
        else:
            S, NS = divmod(float(timestamp or time.time()), 1.0)
            return self.Value(self.type, {
                'value': value,
                'timeStamp': {
                    'secondsPastEpoch': S,
                    'nanoseconds': NS * 1e9,
                },
            })

    typeMap = {
        bool: ntbool,
        int: ntint,
        float: ntfloat,
        unicode: ntstr,
        numpy.ndarray: ntnumericarray.build,
        list: ntstringarray,
    }

    @classmethod
    def unwrap(klass, value):
        """Unpack a Value into an augmented python type (selected from the 'value' field)
        """
        assert isinstance(value, Value), value
        V = value.value
        try:
            T = klass.typeMap[type(V)]
        except KeyError:
            raise ValueError("Can't unwrap value of type %s" % type(V))
        try:
            return T(value.value)._store(value)
        except Exception as e:
            raise ValueError("Can't construct %s around %s (%s): %s" % (T, value, type(value), e))

    def assign(self, V, py):
        """Store python value in Value
        """
        V.value = py

if sys.version_info < (3, 0):
    class ntlong(ntwrappercommon, long):
        pass

    NTScalar.typeMap[long] = ntlong




import logging
import warnings
_log = logging.getLogger(__name__)

from functools import partial
from weakref import WeakSet

import asyncio

from .raw import SharedPV as _SharedPV, Handler
from ..client.raw import LazyRepr
from ..client.thread import RemoteError

__all__ = (
    'SharedPV',
        'Handler',
)

@asyncio.coroutine
def _sync(loop):
    # wait until any pending callbacks are run
    evt = asyncio.Event(loop=loop)
    loop.call_soon(evt.set)
    yield from evt.wait()

    evt.clear() # reuse

    wait4 = set(loop._SharedPV_handlers) # snapshot of in-progress

    # like asyncio.wait() but non-invasive if further callbacks are added.
    # eg. like overlapping calls of _sync()
    # We're abusing the callback chain to avoid creating an Event for
    # each inprogress Future on the assumption that calls to _sync()
    # are relatively rare.

    cnt = len(wait4)
    fut = asyncio.Future(loop=loop)

    def _done(V):
        nonlocal cnt
        cnt -= 1
        if cnt==0 and not fut.cancelled():
            fut.set_result(None)
        return V # pass result along

    if cnt==0:
        fut.set_result(None)
    else:
        [W.add_done_callback(_done) for W in wait4]

    yield from fut

def _log_err(V):
    if isinstance(V, Exception):
        _log.error("Unhandled from SharedPV handler: %s", V)
        # TODO: figure out how to show stack trace...
        # until then, propagate in the hope that someone else will
    return V


def _handle(loop, op, M, args):
    try:
        _log.debug('SERVER HANDLE %s %s %s', op, M, LazyRepr(args))
        maybeco = M(*args)
        if asyncio.iscoroutine(maybeco):
            _log.debug('SERVER SCHEDULE %s', maybeco)
            task = loop.create_task(maybeco)
            task.add_done_callback(_log_err)
            task._log_destroy_pending = False # hack as we don't currently provide a way to join

            loop._SharedPV_handlers.add(task) # track in-progress
        return
    except RemoteError as e:
        err = e
    except Exception as e:
        _log.exception("Unexpected")
        err = e
    if op is not None:
        op.done(error=str(err))


class SharedPV(_SharedPV):

    def __init__(self, handler=None, loop=None, **kws):
        self.loop = loop or asyncio.get_event_loop()
        _SharedPV.__init__(self, handler=handler, **kws)
        self._handler.loop = self.loop
        self._disconnected = asyncio.Event(loop=self.loop)
        self._disconnected.set()
        if not hasattr(self.loop, '_SharedPV_handlers'):
            self.loop._SharedPV_handlers = WeakSet() # holds our in-progress Futures

    def _exec(self, op, M, *args):
        self.loop.call_soon_threadsafe(partial(_handle, self.loop, op, M, args))
        # 3.5 adds asyncio.run_coroutine_threadsafe()

    def _onFirstConnect(self, _junk):
        self._disconnected.clear()

    def _onLastDisconnect(self, _junk):
        self._disconnected.set()

    @asyncio.coroutine
    def _wait_closed(self):
        yield from _sync(self.loop)
        yield from self._disconnected.wait()

    def close(self, destroy=False, sync=False):
        """Close PV, disconnecting any clients.

        :param bool destroy: Indicate "permanent" closure.  Current clients will not see subsequent open().
        :param bool sync: When block until any pending onLastDisconnect() is delivered (timeout applies).
        :param float timeout: Applies only when sync=True.  None for no timeout, otherwise a non-negative floating point value.

        close() with destory=True or sync=True will not prevent clients from re-connecting.
        New clients may prevent sync=True from succeeding.
        Prevent reconnection by __first__ stopping the Server, removing with :py:method:`StaticProvider.remove()`,
        or preventing a :py:class:`DynamicProvider` from making new channels to this SharedPV.
        """
        _SharedPV.close(self, destroy)
        if sync:
            return self._wait_closed()




from __future__ import print_function

import weakref
import gc
import unittest

from ..wrapper import Type as _Type
from .utils import RefTestCase


class TestRawType(RefTestCase):

    def testScalar(self):
        L = [
            ('a', 'i'),
            ('b', 'f'),
        ]
        T = _Type(spec=L)

        self.assertEqual(T.aspy(),
                         ('S', 'structure', L))
        self.assertEqual(T.aspy('a'), 'i')
        self.assertEqual(T['a'], 'i')
        self.assertEqual(len(T), 2)

    def testScalarTest(self):
        L = [
            ('a', 'i'),
            ('b', 'f'),
        ]
        T = _Type(spec=L)

        self.assertTrue(T.has('a'))
        self.assertTrue(T.has('b'))
        self.assertFalse(T.has('c'))
        self.assertListEqual(T.keys(), ['a', 'b'])

        self.assertTrue('a' in T)
        self.assertTrue('b' in T)
        self.assertFalse('c' in T)

    def testID(self):
        L = [
            ('a', 'i'),
            ('b', 'f'),
        ]
        T = _Type(spec=L, id='foo')

        self.assertEqual(T.aspy(),
                         ('S', 'foo', L))

    def testSubStruct(self):
        L = [
            ('a', 'i'),
            ('X', ('S', 'bar', [
                ('m', 's'),
            ])),
            ('b', 'f'),
        ]
        T = _Type(L)

        self.assertEqual(T.aspy(),
                         ('S', 'structure', L))
        self.assertEqual(T['X'].getID(), 'bar')
        self.assertEqual(T['X'].aspy(), T.aspy('X'))
        self.assertEqual(T['X'].aspy(), L[1][1])

    def testUnion(self):
        L = [
            ('a', 'i'),
            ('X', ('U', 'bar', [
                ('m', 's'),
                ('n', 'i'),
            ])),
            ('b', 'f'),
        ]
        T = _Type(L)

        self.assertEqual(T.aspy(),
                         ('S', 'structure', L))

    def testStructArray(self):
        L = [
            ('a', 'i'),
            ('X', ('aS', 'bar', [
                ('m', 's'),
            ])),
            ('b', 'f'),
        ]
        T = _Type(L)

        self.assertEqual(T.aspy(),
                         ('S', 'structure', L))

    def testStructArray2(self):
        L = [
            ('a', 'i'),
            ('X', ('aU', 'bar', [
                ('m', 's'),
                ('n', 'i'),
            ])),
            ('b', 'f'),
        ]
        T = _Type(L)

        self.assertEqual(T.aspy(),
                         ('S', 'structure', L))

    def testAll(self):
        L = [
            ('bool', '?'),
            ('str', 's'),
            ('i8', 'b'),
            ('u8', 'B'),
            ('i16', 'h'),
            ('u16', 'H'),
            ('i32', 'i'),
            ('u32', 'I'),
            ('i64', 'l'),
            ('u64', 'L'),
            ('f32', 'f'),
            ('f64', 'd'),
            ('any', 'v'),
            ('abool', 'a?'),
            ('astr', 'as'),
            ('ai8', 'ab'),
            ('au8', 'aB'),
            ('ai16', 'ah'),
            ('au16', 'aH'),
            ('ai32', 'ai'),
            ('au32', 'aI'),
            ('ai64', 'al'),
            ('au64', 'aL'),
            ('af32', 'af'),
            ('af64', 'ad'),
            ('aany', 'av'),
            ('sub', ('S', 'bar', [
                ('m', 's'),
                ('n', 'i'),
            ])),
            ('asub', ('aS', 'bar', [
                ('m', 's'),
                ('n', 'i'),
            ])),
            ('union', ('U', 'bar', [
                ('m', 's'),
                ('n', 'i'),
            ])),
            ('aunion', ('aU', 'bar', [
                ('m', 's'),
                ('n', 'i'),
            ])),
            ('b', 'f'),
        ]
        T = _Type(L)

        self.assertEqual(T.aspy(),
                         ('S', 'structure', L))

    def testReserved(self):
        L = [
            ("name", "as"),
            ("field", "aI"),
            ("type", "as"),
        ]
        T = _Type(L)

        self.assertEqual(T.aspy(),
                         ('S', 'structure', L))

    def testGC(self):
        T = _Type([('a', 'I')])
        self.assertTrue(gc.is_tracked(T))

        R = weakref.ref(T)
        del T
        gc.collect()
        T = R()
        if T is not None:
            print("Not Dead!", T, gc.get_referrers(T))
        self.assertIsNone(T)

    def testStructID(self):
        T = _Type([('a', 'I')])
        self.assertEqual(T.getID(), "structure")

        T = _Type([('a', 'I')], id="foo")
        self.assertEqual(T.getID(), "foo")

    def testExtend(self):
        B = _Type([('a', 'I')])
        S = _Type([('b', 'I')], base=B)

        self.assertTupleEqual(S.aspy(), ('S', 'structure', [
            ('a', 'I'),
            ('b', 'I'),
        ]))




import logging
import sys
_log = logging.getLogger(__name__)

try:
    from itertools import izip
except ImportError:
    izip = zip

import time
from collections import OrderedDict
from operator import itemgetter
from ..wrapper import Type, Value
from .common import timeStamp, alarm
from .scalar import NTScalar
from .ndarray import NTNDArray
from .enum import NTEnum

__all__ = [
    'NTScalar',
    'NTEnum',
    'NTMultiChannel',
    'NTTable',
    'NTNDArray',
    'defaultNT',
]

_default_nt = {
    "epics:nt/NTScalar:1.0": NTScalar,
    "epics:nt/NTScalarArray:1.0": NTScalar,
    "epics:nt/NTEnum:1.0": NTEnum,
    "epics:nt/NTNDArray:1.0": NTNDArray,
}

def defaultNT():
    """Returns a copy of the default NT helper mappings.

    :since: 3.1.0
    """
    return _default_nt.copy()

class UnwrapOnly(object):
    def __init__(self, unwrap):
        self.unwrap = unwrap
    def __call__(self):
        return self # we are state-less
    def wrap(self, V):
        return V

def buildNT(nt=None, unwrap=None):
    if unwrap is False or nt is False:
        return ClientUnwrapper({}) # disable use of wrappers

    if unwrap is not None:
        # legacy
        ret = {} # ignore new style
        for ID,fn in (unwrap or {}).items():
            ret[ID] = UnwrapOnly(fn)

    else:
        ret = dict(_default_nt)
        ret.update(nt or {})

    return ClientUnwrapper(ret)

class ClientUnwrapper(object):
    def __init__(self, nt=None):
        self.nt = nt
        self.id = None
        self._wrap = self._unwrap = lambda x:x
        self._assign = self._default_assign
    def wrap(self, val):
        """Pack a arbitrary python object into a Value
        """
        return self._wrap(val)
    def unwrap(self, val):
        """Unpack a Value as some other python type
        """
        if val.getID()!=self.id:
            self._update(val)
        return self._unwrap(val)

    def assign(self, V, value):
        if V.getID()!=self.id:
            self._update(V)
        self._assign(V, value)

    def _update(self, val):
        # type change
        nt = self.nt.get(val.getID())
        if nt is not None:
            nt = nt() # instancate
            self._wrap, self._unwrap = nt.wrap, nt.unwrap
            self._assign = nt.assign
            self.id = val.getID()
        else:
            # reset
            self._wrap = self._unwrap = lambda x:x
            self._assign = self._default_assign

    def _default_assign(self, V, value):
        V.value = value # assume NTScalar-like

    def __repr__(self):
        return '%s(%s)'%(self.__class__.__name__, repr(self.nt))
    __str__ = __repr__

class NTMultiChannel(object):

    """Describes a structure holding the equivalent of a number of NTScalar
    """
    Value = Value

    @staticmethod
    def buildType(valtype, extra=[]):
        """Build a Type

        :param str valtype: A type code to be used with the 'value' field.  Must be an array
        :param list extra: A list of tuples describing additional non-standard fields
        :returns: A :py:class:`Type`
        """
        assert valtype[:1] == 'a', 'valtype must be an array'
        return Type(id="epics:nt/NTMultiChannel:1.0",
                    spec=[
                    ('value', valtype),
                    ('channelName', 'as'),
                    ('descriptor', 's'),
                    ('alarm', alarm),
                    ('timeStamp', timeStamp),
                    ('severity', 'ai'),
                    ('status', 'ai'),
                    ('message', 'as'),
                    ('secondsPastEpoch', 'al'),
                    ('nanoseconds', 'ai'),
                    ('userTag', 'ai'),
                    ('isConnected', 'a?'),
                    ] + extra)


class NTTable(object):

    """A generic table

    >>> table = NTTable.buildType(columns=[
        ('columnA', 'ai'),
        ('columnB', 'as'),
    ])
    """
    Value = Value

    @staticmethod
    def buildType(columns=[], extra=[]):
        """Build a table

        :param list columns: List of column names and types. eg [('colA', 'd')]
        :param list extra: A list of tuples describing additional non-standard fields
        :returns: A :py:class:`Type`
        """
        return Type(id="epics:nt/NTTable:1.0",
                    spec=[
                    ('labels', 'as'),
                    ('value', ('S', None, columns)),
                    ('descriptor', 's'),
                    ('alarm', alarm),
                    ('timeStamp', timeStamp),
                    ] + extra)

    def __init__(self, columns=[], extra=[]):
        self.labels = []
        C = []
        for col, type in columns:
            if type[0] == 'a':
                raise ValueError("NTTable column types may not be array")
            C.append((col, 'a' + type))
            self.labels.append(col)
        self.type = self.buildType(C, extra=extra)

    def wrap(self, values):
        """Pack an iterable of dict into a Value

        >>> T=NTTable([('A', 'ai'), ('B', 'as')])
        >>> V = T.wrap([
            {'A':42, 'B':'one'},
            {'A':43, 'B':'two'},
        ])
        """
        if isinstance(values, Value):
            return values
        cols = dict([(L, []) for L in self.labels])
        try:
            # unzip list of dict
            for V in values:
                for L in self.labels:
                    try:
                        cols[L].append(V[L])
                    except (IndexError, KeyError):
                        pass
            # allow omit empty columns
            for L in self.labels:
                V = cols[L]
                if len(V) == 0:
                    del cols[L]

            try:
                return self.Value(self.type, {
                    'labels': self.labels,
                    'value': cols,
                })
            except:
                _log.error("Failed to encode '%s' with %s", cols, self.labels)
                raise
        except:
            _log.exception("Failed to wrap: %s", values)
            raise

    @staticmethod
    def unwrap(value):
        """Iterate an NTTable

        :returns: An iterator yielding an OrderedDict for each column
        """
        ret = []

        # build lists of column names, and value
        lbl, cols = [], []
        for cname, cval in value.value.items():
            lbl.append(cname)
            cols.append(cval)

        # zip together column arrays to iterate over rows
        for rval in izip(*cols):
            # zip together column names and row values
            ret.append(OrderedDict(zip(lbl, rval)))

        return ret

class NTURI(object):

    @staticmethod
    def buildType(args):
        """Build NTURI

        :param list args: A list of tuples of query argument name and PVD type code.

        >>> I = NTURI([
            ('arg_a', 'I'),
            ('arg_two', 's'),
        ])
        """
        try:
            return Type(id="epics:nt/NTURI:1.0", spec=[
                ('scheme', 's'),
                ('authority', 's'),
                ('path', 's'),
                ('query', ('S', None, args)),
            ])
        except Exception as e:
            raise ValueError('Unable to build NTURI compatible type from %s' % args)

    def __init__(self, args):
        self._args = list(args)
        self.type = self.buildType(args)

    def wrap(self, path, args=(), kws={}, scheme='', authority=''):
        """Wrap argument values (tuple/list with optional dict) into Value

        :param str path: The PV name to which this call is made
        :param tuple args: Ordered arguments
        :param dict kws: Keyword arguments
        :rtype: Value
        """
        # build dict of argument name+value
        AV = {}
        AV.update([A for A in kws.items() if A[1] is not None])
        AV.update([(N, V) for (N, _T), V in zip(self._args, args)])

        # list of argument name+type tuples for which a value was provided
        AT = [A for A in self._args if A[0] in AV]

        T = self.buildType(AT)
        try:
            return Value(T, {
                'scheme': scheme,
                'authority': authority,
                'path': path,
                'query': AV,
            })
        except Exception as e:
            raise ValueError('Unable to initialize NTURI %s from %s using %s' % (AT, AV, self._args))




import logging
import warnings
_log = logging.getLogger(__name__)

from functools import partial

from threading import Thread

from .._p4p import SharedPV as _SharedPV
from ..client.raw import LazyRepr

__all__ = (
    'SharedPV',
        'Handler',
)


class ServOpWrap(object):

    def __init__(self, op, unwrap):
        self._op, self._unwrap = op, unwrap

    def value(self):
        return self._unwrap(self._op.value())

    def __getattr__(self, key):
        return getattr(self._op, key)


class Handler(object):

    """Skeleton of SharedPV Handler

    Use of this as a base class is optional.
    """

    def put(self, pv, op):
        """
        Called each time a client issues a Put
        operation on this Channel.

        :param SharedPV pv: The :py:class:`SharedPV` which this Handler is associated with.
        :param ServerOperation op: The operation being initiated.
        """
        op.done(error='Not supported')

    def rpc(self, pv, op):
        """
        Called each time a client issues a Remote Procedure Call
        operation on this Channel.

        :param SharedPV pv: The :py:class:`SharedPV` which this Handler is associated with.
        :param ServerOperation op: The operation being initiated.
        """
        op.done(error='Not supported')

    def onFirstConnect(self, pv):
        """
        Called when the first Client channel is created.

        :param SharedPV pv: The :py:class:`SharedPV` which this Handler is associated with.
        """
        pass

    def onLastDisconnect(self, pv):
        """
        Called when the last Client channel is closed.

        :param SharedPV pv: The :py:class:`SharedPV` which this Handler is associated with.
        """
        pass


class SharedPV(_SharedPV):

    """Shared state Process Variable.  Callback based implementation.

    .. note:: if initial=None, the PV is initially **closed** and
              must be :py:meth:`open()`'d before any access is possible.

    :param handler: A object which will receive callbacks when eg. a Put operation is requested.
                    May be omitted if the decorator syntax is used.
    :param Value initial: An initial Value for this PV.  If omitted, :py:meth:`open()`s must be called before client access is possible.
    :param nt: An object with methods wrap() and unwrap().  eg :py:class:`p4p.nt.NTScalar`.
    :param callable wrap: As an alternative to providing 'nt=', A callable to transform Values passed to open() and post().
    :param callable unwrap: As an alternative to providing 'nt=', A callable to transform Values returned Operations in Put/RPC handlers.
    :param dict options: A dictionary of configuration options.

    Creating a PV in the open state, with no handler for Put or RPC (attempts will error). ::

        from p4p.nt import NTScalar
        pv = SharedPV(nt=NTScalar('d'), value=0.0)
        # ... later
        pv.post(1.0)

    The full form of a handler object is: ::

        class MyHandler:
            def put(self, pv, op):
                pass
            def rpc(self, pv, op):
                pass
            def onFirstConnect(self): # may be omitted
                pass
            def onLastDisconnect(self): # may be omitted
                pass
    pv = SharedPV(MyHandler())

    Alternatively, decorators may be used. ::

        pv = SharedPV()
        @pv.put
        def onPut(pv, op):
            pass

    The nt= or wrap= and unwrap= arguments can be used as a convience to allow
    the open(), post(), and associated Operation.value() to be automatically
    transform to/from :py:class:`Value` and more convienent Python types.
    See :ref:`unwrap`
    """

    def __init__(self, handler=None, initial=None,
                 nt=None, wrap=None, unwrap=None, **kws):
        self.nt = nt
        self._handler = handler or self._DummyHandler()
        self._whandler = self._WrapHandler(self, self._handler)

        self._wrap = wrap or (nt and nt.wrap) or (lambda x: x)
        self._unwrap = unwrap or (nt and nt.unwrap) or (lambda x: x)

        _SharedPV.__init__(self, self._whandler, **kws)
        if initial is not None:
            self.open(initial, nt=nt, wrap=wrap, unwrap=unwrap)

    def open(self, value, nt=None, wrap=None, unwrap=None):
        """Mark the PV as opened an provide its initial value.
        This initial value is later updated with post().

        :param value:  A Value, or appropriate object (see nt= and wrap= of the constructor).

        Any clients which have begun connecting which began connecting while
        this PV was in the close'd state will complete connecting.

        Only those fields of the value which are marked as changed will be stored.
        """

        self._wrap = wrap or (nt and nt.wrap) or self._wrap
        self._unwrap = unwrap or (nt and nt.unwrap) or self._unwrap

        _SharedPV.open(self, self._wrap(value))

    def post(self, value):
        """Provide an update to the Value of this PV.

        :param value:  A Value, or appropriate object (see nt= and wrap= of the constructor).

        Only those fields of the value which are marked as changed will be stored.
        """
        _SharedPV.post(self, self._wrap(value))

    def current(self):
        return self._unwrap(_SharedPV.current(self))

    def _exec(self, op, M, *args):  # sub-classes will replace this
        try:
            M(*args)
        except Exception as e:
            if op is not None:
                op.done(error=str(e))
            _log.exception("Unexpected")

    def _onFirstConnect(self, _junk):
        pass # see sub-classes.  run before user onFirstConnect()

    def _onLastDisconnect(self, _junk):
        pass # see sub-classes.  run after user onLastDisconnect()

    class _DummyHandler(object):
        pass

    class _WrapHandler(object):

        "Wrapper around user Handler which logs exceptions"

        def __init__(self, pv, real):
            self._pv = pv  # this creates a reference cycle, which should be collectable since SharedPV supports GC
            self._real = real

        def onFirstConnect(self):
            self._pv._exec(None, self._pv._onFirstConnect, None)
            try:  # user handler may omit onFirstConnect()
                M = self._real.onFirstConnect
            except AttributeError:
                return
            self._pv._exec(None, M, self._pv)

        def onLastDisconnect(self):
            try:
                M = self._real.onLastDisconnect
            except AttributeError:
                pass
            else:
                self._pv._exec(None, M, self._pv)
            self._pv._exec(None, self._pv._onLastDisconnect, None)

        def put(self, op):
            _log.debug('PUT %s %s', self._pv, op)
            try:
                self._pv._exec(op, self._real.put, self._pv, ServOpWrap(op, self._pv._unwrap))
            except AttributeError:
                op.done(error="Put not supported")

        def rpc(self, op):
            _log.debug('RPC %s %s', self._pv, op)
            try:
                self._pv._exec(op, self._real.rpc, self._pv, op)
            except AttributeError:
                op.done(error="RPC not supported")

    @property
    def onFirstConnect(self):
        def decorate(fn):
            self._handler.onFirstConnect = fn
            return fn
        return decorate

    @property
    def onLastDisconnect(self):
        def decorate(fn):
            self._handler.onLastDisconnect = fn
            return fn
        return decorate

    @property
    def put(self):
        def decorate(fn):
            self._handler.put = fn
            return fn
        return decorate

    @property
    def rpc(self):
        def decorate(fn):
            self._handler.rpc = fn
            return fn
        return decorate

    def __repr__(self):
        if self.isOpen():
            return '%s(value=%s)' % (self.__class__.__name__, repr(self.current()))
        else:
            return "%s(<closed>)" % (self.__class__.__name__,)
    __str__ = __repr__




from __future__ import absolute_import

import logging
import warnings
_log = logging.getLogger(__name__)

from weakref import WeakSet
from functools import partial

import cothread
from cothread import Spawn, Callback, Event, WaitForAll
from .raw import SharedPV as _SharedPV, Handler
from ..client.thread import RemoteError

__all__ = (
    'SharedPV',
    'Handler',
)

# Spawn and Event have __slots__ which doesn't include '__weakref__'

class WeakSpawn(Spawn):
    __slots__=['__weakref__']

class WeakEvent(Event):
    __slots__=['__weakref__']

# a set of WeakSpawn and/or WeakEvent
# in progress _handle() cothreads, or _sync() events
_handlers = WeakSet()

def _sync(timeout=None):
    """I will wait until all pending handlers cothreads have completed
    """
    evt = WeakEvent(auto_reset=False)

    # first ensure that any pending callbacks from worker threads have been delivered
    # these are calls of _fromMain()
    Callback(evt.Signal)
    evt.Wait(timeout=timeout)

    evt.Reset() # reuse

    # grab the current set of inprogress cothreads/events
    wait4 = set(_handlers)
    # because Spawn.Wait() can only be called once, remove them and
    # use 'evt' as a proxy for what I'm waiting on so that overlapping
    # calls to _sync() will wait for these as well.
    # However, this means that our failure will must cascade to subsequent
    # calls to _sync() before we complete.
    _handlers.clear()
    _handlers.add(evt)

    try:
        WaitForAll(wait4, timeout=timeout)
    except Exception as e:
        evt.SignalException(e) # pass along error to next concurrent _sync()
    else:
        evt.Signal() # pass along success

# Callback() runs me in the main thread
def _fromMain(_handle, op, M, args):
    #_log.debug("_fromMain %s", M)
    co = WeakSpawn(_handle, op, M, args)
    _handlers.add(co)

def _handle(op, M, args):
    try:
        M(*args)
        return
    except RemoteError as e:
        err = e
    except Exception as e:
        _log.exception("Unexpected")
        err = e
    finally:
        M = args = None # prevent lingering references after _sync() returns
    if op is not None:
        op.done(error=str(err))


class SharedPV(_SharedPV):

    def __init__(self, queue=None, **kws):
        _SharedPV.__init__(self, **kws)
        self._queue = queue or Callback
        self._disconnected = Event(auto_reset=False)
        self._disconnected.Signal()

    def _exec(self, op, M, *args):
        #_log.debug("_exec %s", M)
        self._queue(_fromMain, _handle, op, M, args)

    def _onFirstConnect(self, _junk):
        self._disconnected.Reset()

    def _onLastDisconnect(self, _junk):
        self._disconnected.Signal()

    def close(self, destroy=False, sync=False, timeout=None):
        """Close PV, disconnecting any clients.

        :param bool destroy: Indicate "permanent" closure.  Current clients will not see subsequent open().
        :param bool sync: When block until any pending onLastDisconnect() is delivered (timeout applies).
        :param float timeout: Applies only when sync=True.  None for no timeout, otherwise a non-negative floating point value.

        close() with destory=True or sync=True will not prevent clients from re-connecting.
        New clients may prevent sync=True from succeeding.
        Prevent reconnection by __first__ stopping the Server, removing with :py:method:`StaticProvider.remove()`,
        or preventing a :py:class:`DynamicProvider` from making new channels to this SharedPV.
        """
        _SharedPV.close(self, destroy)
        if sync:
            # TODO: still not syncing PVA workers...
            _sync()
            self._disconnected.Wait(timeout=timeout)




import sys
import logging
import inspect
from functools import wraps, partial
_log = logging.getLogger(__name__)

from threading import Thread

from .wrapper import Value, Type
from .nt import NTURI
from .client.raw import RemoteError, LazyRepr
from .server import DynamicProvider
from .server.raw import SharedPV
from .util import ThreadedWorkQueue, WorkQueue, Full, Empty

__all__ = [
    'rpc',
    'rpccall',
    'rpcproxy',
    'RemoteError',
    'WorkQueue',
    'NTURIDispatcher',
]


def rpc(rtype=None):
    """Decorator marks a method for export.

    :param type: Specifies which :py:class:`Type` this method will return.

    The return type (rtype) must be one of:

    - An instance of :py:class:`p4p.Type`
    - None, in which case the method must return a :py:class:`p4p.Value`
    - One of the NT helper classes (eg :py:class:`p4p.nt.NTScalar`).
    - A list or tuple used to construct a :py:class:`p4p.Type`.

    Exported methods raise an :py:class:`Exception` to indicate an error to the remote caller.
    :py:class:`RemoteError` may be raised to send a specific message describing the error condition.

    >>> class Example(object):
        @rpc(NTScalar.buildType('d'))
        def add(self, lhs, rhs):
            return {'value':float(lhs)+flost(rhs)}
    """
    wrap = None
    if rtype is None or isinstance(rtype, Type):
        pass
    elif isinstance(type, (list, tuple)):
        rtype = Type(rtype)
    elif hasattr(rtype, 'type'):  # eg. one of the NT* helper classes
        wrap = rtype.wrap
        rtype = rtype.type
    else:
        raise TypeError("Not supported")

    def wrapper(fn):
        if wrap is not None:
            orig = fn

            @wraps(orig)
            def wrapper2(*args, **kws):
                return wrap(orig(*args, **kws))
            fn = wrapper2

        fn._reply_Type = rtype
        return fn
    return wrapper


def rpccall(pvname, request=None, rtype=None):
    """Decorator marks a client proxy method.

    :param str pvname: The PV name, which will be formated using the 'format' argument of the proxy class constructor.
    :param request: A pvRequest string or :py:class:`p4p.Value` passed to eg. :py:meth:`p4p.client.thread.Context.rpc`.

    The method to be decorated must have all keyword arguments,
    where the keywords are type code strings or :class:`~p4p.Type`.

    """
    def wrapper(fn):
        fn._call_PV = pvname
        fn._call_Request = request
        fn._reply_Type = rtype
        return fn
    return wrapper


class RPCDispatcherBase(DynamicProvider):
    def __init__(self, queue, target=None, channels=set(), name=None):
        DynamicProvider.__init__(self, name, self)  # we are our own Handler
        self.queue = queue
        self.target = target
        self.channels = set(channels)
        self.name = name
        self.__pv = SharedPV(
            handler=self,  # no per-channel state, and only RPC used, so only need on PV
            initial=Value(Type([])),  # we don't support get/put/monitor, so use empty struct
        )
        M = self.methods = {}
        for name, mem in inspect.getmembers(target):
            if not hasattr(mem, '_reply_Type'):
                continue
            M[name] = mem

    def getMethodNameArgs(self, request):
        raise NotImplementedError("Sub-class must implement getMethodName")
        # sub-class needs to extract method name from request
        # return 'name', {'var':'val'}

    def testChannel(self, name):
        _log.debug("Test RPC channel %s = %s", name, name in self.channels)
        return name in self.channels

    def makeChannel(self, name, src):
        if self.testChannel(name):
            _log.debug("Open RPC channel %s", name)
            return self.__pv  # no per-channel tracking needed
        else:
            _log.warn("Ignore RPC channel %s", name)

    def rpc(self, pv, op):
        _log.debug("RPC call %s", op)
        try:
            self.queue.push(partial(self._handle, op))
        except Full:
            _log.warn("RPC call queue overflow")
            op.done(error="Too many concurrent RPC calls")

    def _handle(self, op):
        try:
            request = op.value()
            name, args = self.getMethodNameArgs(request)
            fn = self.methods[name]
            rtype = fn._reply_Type

            R = fn(**args)

            if not isinstance(R, Value):
                try:
                    R = Value(rtype, R)
                except:
                    _log.exception("Error encoding %s as %s", R, rtype)
                    op.done(error="Error encoding reply")
                    return
            _log.debug("RPC reply %s -> %s", request, LazyRepr(R))
            op.done(R)

        except RemoteError as e:
            _log.debug("RPC reply %s -> error: %s", request, e)
            op.done(error=str(e))

        except:
            _log.exception("Error handling RPC %s", request)
            op.done(error="Error handling RPC")


class NTURIDispatcher(RPCDispatcherBase):

    """RPC dispatcher using NTURI (a al. eget)

    Method names are prefixed with a fixed string.

    >>> queue = WorkQueue()
    >>> class Summer(object):
        @rpc([('result', 'i')])
        def add(self, a=None, b=None):
            return {'result': int(a)+int(b)}
    >>> installProvider("arbitrary", NTURIDispatcher(queue, target=Summer(), prefix="pv:prefix:"))

    Making a call with the CLI 'eget' utility::

      $ eget -s pv:prefix:add -a a=1 -a b=2
      ....
      int result 3

    :param queue WorkQueue: A WorkQueue to which RPC calls will be added
    :param prefix str: PV name prefix used by RPC methods
    :param target: The object which has the RPC calls
    """

    def __init__(self, queue, prefix=None, **kws):
        RPCDispatcherBase.__init__(self, queue, **kws)
        self.prefix = prefix
        self.methods = dict([(prefix + meth, fn) for meth, fn in self.methods.items()])
        self.channels = set(self.methods.keys())
        _log.debug('NTURI methods: %s', ', '.join(self.channels))

    def getMethodNameArgs(self, request):
        # {'schema':'pva', 'path':'pvname', 'query':{'var':'val', ...}}
        return request.path, dict(request.query.items())

# legecy for MASAR only
# do not use in new code


class MASARDispatcher(RPCDispatcherBase):

    def __init__(self, queue, **kws):
        RPCDispatcherBase.__init__(self, queue, **kws)
        _log.debug("MASAR pv %s methods %s", self.channels, self.methods)

    def getMethodNameArgs(self, request):
        # all through a single PV, method name in request
        # {'function':'rpcname', 'name':['name', ...], 'value':['val', ...]}
        return request.function, dict(zip(request.get('name', []), request.get('value', [])))


def quickRPCServer(provider, prefix, target,
                   maxsize=20,
                   workers=1,
                   useenv=True, conf=None, isolate=False):
    """Run an RPC server in the current thread

    Calls are handled sequentially, and always in the current thread, if workers=1 (the default).
    If workers>1 then calls are handled concurrently by a pool of worker threads.
    Requires NTURI style argument encoding.

    :param str provider: A provider name.  Must be unique in this process.
    :param str prefix: PV name prefix.  Along with method names, must be globally unique.
    :param target: The object which is exporting methods.  (use the :func:`rpc` decorator)
    :param int maxsize: Number of pending RPC calls to be queued.
    :param int workers: Number of worker threads (default 1)
    :param useenv: Passed to :class:`~p4p.server.Server`
    :param conf: Passed to :class:`~p4p.server.Server`
    :param isolate: Passed to :class:`~p4p.server.Server`
    """
    from p4p.server import Server
    import time
    queue = ThreadedWorkQueue(maxsize=maxsize, workers=workers)
    provider = NTURIDispatcher(queue, target=target, prefix=prefix, name=provider)
    threads = []
    server = Server(providers=[provider], useenv=useenv, conf=conf, isolate=isolate)
    with server, queue:
        while True:
            time.sleep(10.0)


class RPCProxyBase(object):

    """Base class for automatically generated proxy classes
    """
    context = None
    "The Context provided on construction"
    format = None
    "The tuple/dict used to format ('%' operator) PV name strings."
    timeout = 3.0
    "Timeout of RPC calls in seconds"
    authority = ''
    "Authority string sent with NTURI requests"
    throw = True
    "Whether call errors raise an exception, or return it"
    scheme = None  # set to override automatic


def _wrapMethod(K, V):
    pv, req = V._call_PV, V._call_Request
    if sys.version_info >= (3, 0):
        S = inspect.getfullargspec(V)
        keywords = S.varkw
    else:
        S = inspect.getargspec(V)
        keywords = S.keywords

    if S.varargs is not None or keywords is not None:
        raise TypeError("vararg not supported for proxy method %s" % K)

    if len(S.args) != len(S.defaults):
        raise TypeError("proxy method %s must specify types for all arguments" % K)

    try:
        NT = NTURI(zip(S.args, S.defaults))
    except Exception as e:
        raise TypeError("%s : failed to build method from %s, %s" % (e, S.args, S.defaults))

    @wraps(V)
    def mcall(self, *args, **kws):
        pvname = pv % self.format
        try:
            uri = NT.wrap(pvname, args, kws, scheme=self.scheme or self.context.name, authority=self.authority)
        except Exception as e:
            raise ValueError("Unable to wrap %s %s as %s (%s)" % (args, kws, NT, e))
        return self.context.rpc(pvname, uri, request=req, timeout=self.timeout, throw=self.throw)

    return mcall


def rpcproxy(spec):
    """Decorator to enable this class to proxy RPC client calls

    The decorated class constructor takes two additional arguments,
    `context=` is required to be a :class:`~p4p.client.thread.Context`.
    `format`= can be a string, tuple, or dictionary and is applied
    to PV name strings given to :py:func:`rpcall`.
    Other arguments are passed to the user class constructor. ::

       @rpcproxy
       class MyProxy(object):
           @rpccall("%s:add")
           def add(lhs='d', rhs='d'):
               pass

       ctxt = Context('pva')
       proxy = MyProxy(context=ctxt, format="tst:")  # evaluates "%s:add"%"tst:"

    The decorated class will be a sub-class of the provided class and :class:`RPCProxyBase`.
    """
    # inject our ctor first so we don't have to worry about super() non-sense.

    def _proxyinit(self, context=None, format={}, **kws):
        assert context is not None, context
        self.context = context
        self.format = format
        spec.__init__(self, **kws)
    obj = {'__init__': _proxyinit}

    for K, V in inspect.getmembers(spec, lambda M: hasattr(M, '_call_PV')):
        obj[K] = _wrapMethod(K, V)

    return type(spec.__name__, (RPCProxyBase, spec), obj)



from ply import lex

class ACFError(RuntimeError):
    pass

tokens = (
    'UAG',
    'HAG',
    'ASG',
    'RULE',
    'CALC',
    'INP',
    'INTEGER',
    'STRING',
)

literals = ('(', ')', '{', '}', ',')

t_ignore  = ' \t\r'

def t_eol(t):
    r'\n+'
    t.lexer.lineno += len(t.value)

def t_comment(t):
    r'\#[^\n]*\n'
    t.lexer.lineno += 1

def t_KW(t):
    r'UAG|HAG|ASG|RULE|CALC'
    t.type = t.value
    return t

def t_INP(t):
    r'INP[A-L]'
    t.type = 'INP'
    return t

def t_INTEGER(t):
    r'[0-9]+'
    t.type = 'INTEGER'
    t.value = int(t.value)
    return t

def t_bare_STRING(t):
    r'[a-zA-Z0-9_\-+:.\[\]<>;]+'
    t.type = 'STRING'
    return t

def t_quoted_STRING(t):
    r'"([^"\n\\]|\\.)*"'
    t.type = 'STRING'
    t.value = t.value[1:-1] # asLib_lex.l doesn't un-escape  .encode('utf-8').decode('unicode_escape')
    return t

def t_error(t):
    raise ACFError("Illegal char %s at %d"%(repr(t.value[:5]), t.lexer.lineno+1))

if __name__=='__main__':
    import logging
    logging.basicConfig(level=logging.DEBUG)
    lexer = lex.lex(debug=1, optimize=0, debuglog=logging.getLogger(__name__))
    lex.runmain(lexer=lexer)



# I'm imported from test_asyncio.y
# Add new TestCase s to __all__

import logging
import warnings
_log = logging.getLogger(__name__)

import unittest
import sys
import random
import weakref
import gc
from unittest.case import SkipTest

from ..nt import NTScalar
from ..server import Server, StaticProvider
from .utils import RefTestCase

import asyncio

from ..client.asyncio import Context, Disconnected, timesout
from ..server.asyncio import SharedPV

from .utils import inloop, clearloop

__all__ = (
    'TestGPM',
    'TestTimeout',
    'TestFirstLast',
)

class Handler:

    @asyncio.coroutine
    def put(self, pv, op):
        _log.debug("putting %s <- %s", op.name(), op.value())
        yield from asyncio.sleep(0, loop=self.loop)  # prove that we can
        pv.post(op.value() * 2)
        op.done()

class TestGPM(RefTestCase):
    timeout = 3  # overall timeout for each test method

    @inloop
    @asyncio.coroutine
    def setUp(self):
        super(TestGPM, self).setUp()

        self.pv = SharedPV(nt=NTScalar('i'), initial=0, handler=Handler(), loop=self.loop)
        self.pv2 = SharedPV(handler=Handler(), nt=NTScalar('d'), initial=42.0, loop=self.loop)
        self.provider = StaticProvider("serverend")
        self.provider.add('foo', self.pv)
        self.provider.add('bar', self.pv2)

    def tearDown(self):
        del self.pv
        del self.pv2
        del self.provider
        clearloop(self)
        super(TestGPM, self).tearDown()

    @inloop
    @asyncio.coroutine
    def test_getput(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False, loop=self.loop) as C:
                self.assertEqual(0, (yield from C.get('foo')))

                yield from C.put('foo', 5)

                self.assertEqual(5 * 2, (yield from C.get('foo')))

    @inloop
    @asyncio.coroutine
    def test_monitor(self):
        with Server(providers=[self.provider], isolate=True) as S:
            with Context('pva', conf=S.conf(), useenv=False, loop=self.loop) as C:

                Q = asyncio.Queue(loop=self.loop)

                sub = C.monitor('foo', Q.put, notify_disconnect=True)
                try:
                    self.assertIsInstance((yield from Q.get()), Disconnected)

                    self.assertEqual(0, (yield from Q.get()))

                    yield from C.put('foo', 2)

                    self.assertEqual(2 * 2, (yield from Q.get()))

                    self.pv.close()

                    self.assertIsInstance((yield from Q.get()), Disconnected)

                    self.pv.open(3)

                    self.assertEqual(3, (yield from Q.get()))

                finally:
                    sub.close()
                    yield from sub.wait_closed()


class TestTimeout(unittest.TestCase):

    def tearDown(self):
        clearloop(self)

    @inloop
    def test_timeout(self):
        done = None

        if sys.version_info >= (3, 4) and sys.version_info < (3, 5):
            raise SkipTest("wait_for() kind of broken in 3.4")
            # I'm seeing a test failure with "got Future <Future pending> attached to a different loop"
            # but I can't find where the different loop gets mixed in.
            # So I _think_ this is a bug in 3.4.

        @timesout()
        @asyncio.coroutine
        def action(loop):
            nonlocal done
            done = False
            yield from asyncio.sleep(5, loop=loop)
            done = True

        try:
            yield from action(self.loop, timeout=0.01)
            self.assertTrue(False)
        except asyncio.TimeoutError:
            pass

        self.assertIs(done, False)

class TestFirstLast(RefTestCase):
    maxDiff = 1000
    timeout = 1.0
    mode = 'Mask'

    class Handler:
        def __init__(self, loop):
            self.evt = asyncio.Event(loop=loop)
            self.conn = None
        def onFirstConnect(self, pv):
            _log.debug("onFirstConnect")
            self.conn = True
            self.evt.set()
        def onLastDisconnect(self, pv):
            _log.debug("onLastDisconnect")
            self.conn = False
            self.evt.set()

    @inloop
    @asyncio.coroutine
    def setUp(self):
        # gc.set_debug(gc.DEBUG_LEAK)
        super(TestFirstLast, self).setUp()

        self.H = self.Handler(self.loop)
        self.pv = SharedPV(handler=self.H,
                           nt=NTScalar('d'),
                           options={'mapperMode':self.mode},
                           loop=self.loop)
        self.sprov = StaticProvider("serverend")
        self.sprov.add('foo', self.pv)

        self.server = Server(providers=[self.sprov], isolate=True)

    def tearDown(self):
        self.server.stop()
        #_defaultWorkQueue.stop()
        del self.server
        del self.sprov
        del self.pv
        del self.H
        clearloop(self)
        super(TestFirstLast, self).tearDown()

    @inloop
    @asyncio.coroutine
    def testClientDisconn(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}, loop=self.loop) as ctxt:
            Q = asyncio.Queue(maxsize=4, loop=self.loop)
            sub = ctxt.monitor('foo', Q.put, notify_disconnect=True)
            try:

                yield from Q.get() # initial update

                _log.debug('TEST')
                yield from self.H.evt.wait() # onFirstConnect()
                self.H.evt.clear()
                self.assertTrue(self.H.conn)

            finally:
                sub.close()
                yield from sub.wait_closed()

        yield from self.H.evt.wait() # onLastDisconnect()
        _log.debug('SHUTDOWN')
        self.assertFalse(self.H.conn)

    @inloop
    @asyncio.coroutine
    def testServerShutdown(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}, loop=self.loop) as ctxt:
            Q = asyncio.Queue(maxsize=4, loop=self.loop)
            sub = ctxt.monitor('foo', Q.put, notify_disconnect=True)
            try:

                yield from Q.get() # initial update

                _log.debug('TEST')
                yield from self.H.evt.wait() # onFirstConnect()
                self.H.evt.clear()
                self.assertIs(self.H.conn, True)

                self.server.stop()

                yield from self.H.evt.wait() # onLastDisconnect()
                _log.debug('SHUTDOWN')
                self.assertIs(self.H.conn, False)

            finally:
                sub.close()
                yield from sub.wait_closed()

    @inloop
    @asyncio.coroutine
    def testPVClose(self):
        self.pv.open(1.0)

        with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}, loop=self.loop) as ctxt:
            Q = asyncio.Queue(maxsize=4, loop=self.loop)
            sub = ctxt.monitor('foo', Q.put, notify_disconnect=True)
            try:

                yield from Q.get() # initial update

                _log.debug('TEST')
                yield from self.H.evt.wait() # onFirstConnect()
                self.H.evt.clear()
                self.assertTrue(self.H.conn)

                yield from self.pv.close(destroy=True, sync=True) # onLastDisconnect()

                _log.debug('CLOSE')
                self.assertFalse(self.H.conn)

            finally:
                sub.close()
                yield from sub.wait_closed()



import logging
import warnings
import weakref
import collections
import gc

import unittest

from ..asLib import Engine
from ..asLib.yacc import parse as parse_acf
from ..asLib.pvlist import PVList, _sub_add

from .. import nt

_log = logging.getLogger(__name__)

class TestPVList(unittest.TestCase):
    def test_slac(self):
        pvlist = r"""
EVALUATION ORDER ALLOW, DENY
# comment
.* ALLOW 
OTRS:DMP1:695:Image:.*     DENY
PATT:SYS0:1:MPSBURSTCTRL.* ALLOW CANWRITE
PATT:SYS0:1:MPSBURSTCTRL.* DENY FROM 1.2.3.4 
PATT:SYS0:1:MPSBURSTCTRL.* DENY From 4.3.2.1
X(.*) ALIAS Y\1 CANWRITE
BEAM.* DENY FROM 1.2.3.4   
BEAM.* ALLOW
BEAM.* ALLOW RWINSTRMCC 1
BEAM:L1:.* allow

THIS ALIAS THAT

a:([^:]*):b:([^:]*) ALIAS A:\2:B:\1

"""

        pvl = PVList(pvlist)

        self.assertEqual(pvl.compute(b'BEAM:stuff', '127.0.0.1'), ('BEAM:stuff', 'RWINSTRMCC', 1))
        self.assertEqual(pvl.compute(b'BEAM:L1:Energy', '127.0.0.1'), ('BEAM:L1:Energy', 'DEFAULT', 0))

        self.assertEqual(pvl.compute(b'OTHER:stuff', '127.0.0.1'), ('OTHER:stuff', 'DEFAULT', 0))

        self.assertEqual(pvl.compute(b'OTRS:DMP1:695:Image:X', '127.0.0.1'), (None, None, None))

        self.assertEqual(pvl.compute(b'PATT:SYS0:1:MPSBURSTCTRLX', '127.0.0.1'), ('PATT:SYS0:1:MPSBURSTCTRLX', 'CANWRITE', 0))
        self.assertEqual(pvl.compute(b'PATT:SYS0:1:MPSBURSTCTRLX', '1.2.3.4'), (None, None, None))
        self.assertEqual(pvl.compute(b'PATT:SYS0:1:MPSBURSTCTRLX', '4.3.2.1'), (None, None, None))

        self.assertEqual(pvl.compute(b'Xsomething', '127.0.0.1'), ('Ysomething', 'CANWRITE', 0))
        self.assertEqual(pvl.compute(b'a:one:b:two', '127.0.0.1'), ('A:two:B:one', 'DEFAULT', 0))

class LocalSubscription(object):
    def __init__(self, cb, code='d'):
        self.cb = cb
        self.nt = nt.NTScalar(code)
    def close(self):
        self.cb = None
    def post(self, V):
        if V is not None:
            V = self.nt.unwrap(self.nt.wrap(V))
        cb = self.cb
        if cb is not None:
            cb(V)

class LocalContext(object):
    def __init__(self, provider):
        assert provider=='pva', provider
        self._sub = collections.defaultdict(weakref.WeakSet)

    def monitor(self, pv, cb, notify_disconnect=False):
        assert notify_disconnect is True
        S = LocalSubscription(cb)
        S.post(None)
        self._sub[pv].add(S)
        return S

    def post(self, pv, V):
        for S in set(self._sub[pv]):
            S.post(V)

class DummyEngine(Engine):
    @staticmethod
    def _gethostbyname(host):
        return {
            'localhost':'127.0.0.1',
            'lcls-daemon3':'1.2.3.4',
            'hxr-control':'1.2.3.44',
            'hxr-console':'1.2.3.45',
            'mcrhost':'1.2.3.10',
            'remotehost':'1.2.3.20',
        }.get(host)

class TestACF(unittest.TestCase):
    def test_parse(self):
        inp='''
UAG(SPECIAL) {
    root,
    "role/admin"
}
HAG(GWSTATS)
{
        lcls-daemon3,
        other.host,
        "strange",
        "invalid.host.name."
}
ASG(SIMPLE)
{
        RULE(1,READ)
}
ASG(NOTSIMPLE) {
  INPA("ACC-CT{}Prmt:Remote-Sel")
        RULE(1,READ)
        RULE(1,WRITE,TRAPWRITE) {
                UAG(PHOTON)
                HAG(PHOTON, OTHER)
    CALC("A!=0")
                
        }
}
'''

        ast = parse_acf(inp)

        self.assertEqual(ast, [
            ('UAG', 'SPECIAL', ['root', 'role/admin']),
            ('HAG', 'GWSTATS', ['lcls-daemon3', 'other.host', 'strange', 'invalid.host.name.']),
            ('ASG', 'SIMPLE', [
                ('RULE', 1, 'READ', False, None)
            ]),
            ('ASG', 'NOTSIMPLE', [
                ('INP', 'A', 'ACC-CT{}Prmt:Remote-Sel'),
                ('RULE', 1, 'READ', False, None),
                ('RULE', 1, 'WRITE', True, [
                    ('UAG', ['PHOTON']),
                    ('HAG', ['PHOTON', 'OTHER']),
                    ('CALC', 'A!=0'),
                ]),
            ]),
        ])

class TestACL(unittest.TestCase):
    class DummyChannel(object):
        def __init__(self):
            self.perm = None
        def access(self, **kws):
            self.perm = kws

    def tearDown(self):
        # cleanup lingering refs. to Type arising through LocalSubscription
        # which would cause false positive failures in subsequent cases
        gc.collect()
        super(TestACL, self).tearDown()

    def test_default(self):
        eng = DummyEngine()
        self.assertIsNone(eng._ctxt)

        ch = self.DummyChannel()
        eng.create(ch, 'DEFAULT', 'someone', 'somewhere', 0)
        self.assertDictEqual(ch.perm, {'put':True, 'rpc':True, 'uncached':True, 'audit': False})

        ch = self.DummyChannel()
        eng.create(ch, 'othergrp', 'someone', 'somewhere', 0)
        self.assertDictEqual(ch.perm, {'put':True, 'rpc':True, 'uncached':True, 'audit': False})

    def test_roles(self):
        eng = DummyEngine("""
UAG(SPECIAL) {
    root,
    "role/admin"
}
ASG(DEFAULT)
{
        RULE(1,READ)
        RULE(1,WRITE) {
            UAG(SPECIAL)
        }
}
""")

        for args, perm in [(('DEFAULT', 'someone', 'somewhere', 0),          {'put':False,'rpc':False, 'uncached':False, 'audit': False}),
                           (('DEFAULT', 'root', '1.2.3.4', 0),               {'put':True, 'rpc':True, 'uncached':False, 'audit': False}),
                           (('DEFAULT', 'someone', '1.2.3.4', 0, ['admin']), {'put':True, 'rpc':True, 'uncached':False, 'audit': False}),
                           ]:
            try:
                _log.debug('With: %s expect: %s', args, perm)
                ch = self.DummyChannel()
                eng.create(ch, *args)
                self.assertDictEqual(ch.perm, perm)
            except AssertionError as e:
                raise AssertionError('%s -> %s : %s'%(args, perm ,e))

    def test_slac(self):
        eng = DummyEngine("""
UAG(PHOTON)
{
        root
}
UAG(tst-opr)
{
        tstioc, tstopr
}
HAG(GWSTATS)
{
        lcls-daemon3
}
HAG(PHOTON)
{
        hxr-control, hxr-console,
        "invalid.host.name."
}
ASG(DEFAULT)
{
        RULE(1,READ)
}
ASG(CANWRITE)
{
        RULE(1,READ)  
        RULE(1,WRITE,TRAPWRITE)
}
ASG(AMOWRITE)
{
        RULE(1,READ)
        RULE(1,WRITE,TRAPWRITE)
                {
                UAG(PHOTON)
                HAG(PHOTON)
                }
}
ASG(TSTWRITE)
{
        RULE(1,READ)
        RULE(1,WRITE,TRAPWRITE)
                {
                UAG( PHOTON , tst-opr )
                HAG(PHOTON,GWSTATS)
                }
}
""")

        for args, perm in [(('DEFAULT', 'someone', 'somewhere', 0),  {'put':False, 'rpc':False, 'uncached':False, 'audit': False}),
                           (('DEFAULT', 'root', '1.2.3.4', 0),       {'put':False, 'rpc':False, 'uncached':False, 'audit': False}),
                           (('CANWRITE', 'someone', 'somewhere', 0), {'put':True , 'rpc':True,  'uncached':False, 'audit': True}),
                           (('CANWRITE', 'root', '1.2.3.44', 0),     {'put':True , 'rpc':True,  'uncached':False, 'audit': True}),
                           (('AMOWRITE', 'someone', 'somewhere', 0), {'put':False, 'rpc':False, 'uncached':False, 'audit': False}),
                           (('AMOWRITE', 'someone', '1.2.3.44', 0),  {'put':False, 'rpc':False, 'uncached':False, 'audit': False}),
                           (('AMOWRITE', 'root', 'somewhere', 0),    {'put':False, 'rpc':False, 'uncached':False, 'audit': False}),
                           (('AMOWRITE', 'root', '1.2.3.44', 0),     {'put':True , 'rpc':True,  'uncached':False, 'audit': True}),
                           ]:
            try:
                ch = self.DummyChannel()
                eng.create(ch, *args)
                self.assertDictEqual(ch.perm, perm)
            except AssertionError as e:
                raise AssertionError('%s -> %s : %s'%(args, perm ,e))

    def test_bnl(self):
        eng = DummyEngine("""
HAG(mcr) {mcrhost}
UAG(softioc) {softioc, rtems}
HAG(remote) {remotehost}
ASG(OPERATOR) {
  INPA("ACC-CT{}Prmt:Remote-Sel")

  RULE(1, READ)

  RULE(1, WRITE, TRAPWRITE) {
    HAG(mcr)
  }

# Allow inter-IOC and physics app writes
  RULE(1, WRITE, TRAPWRITE) {
    UAG(softioc)
  }

# Conditionally allow remote consoles
  RULE(1, WRITE, TRAPWRITE) {
    HAG(remote)
    CALC("A!=0")
  }
}
""", ctxt=LocalContext('pva'))
        self.assertIsNotNone(eng._ctxt)

        db = []
        for args in [('OPERATOR', 'joe', '1.2.3.10', 0),     # always allowed MCR
                     ('OPERATOR', 'softioc', '1.2.3.99', 0), # always allowed IOC
                     ('OPERATOR', 'joe', '1.2.3.99', 0),     # never allowed
                     ('OPERATOR', 'joe', '1.2.3.20', 0),     # conditional
                     ]:
            ch = self.DummyChannel()
            eng.create(ch, *args)
            db.append(ch)

        for ch, perm in zip(db, [{'put':True , 'rpc':True, 'uncached':False, 'audit': True},
                                 {'put':True , 'rpc':True, 'uncached':False, 'audit': True},
                                 {'put':False , 'rpc':False, 'uncached':False, 'audit': False},
                                 {'put':False , 'rpc':False, 'uncached':False, 'audit': False},
                                 ]):
            try:
                self.assertDictEqual(ch.perm, perm)
            except AssertionError as e:
                raise AssertionError('%s : %s'%(perm ,e))

        eng._ctxt.post('ACC-CT{}Prmt:Remote-Sel', 1.0)

        for ch, perm in zip(db, [{'put':True , 'rpc':True, 'uncached':False, 'audit': True},
                                 {'put':True , 'rpc':True, 'uncached':False, 'audit': True},
                                 {'put':False , 'rpc':False, 'uncached':False, 'audit': False},
                                 {'put':True , 'rpc':True, 'uncached':False, 'audit': True},
                                 ]):
            try:
                self.assertDictEqual(ch.perm, perm)
            except AssertionError as e:
                raise AssertionError('%s : %s'%(perm ,e))

        eng._ctxt.post('ACC-CT{}Prmt:Remote-Sel', 0.0)

        for ch, perm in zip(db, [{'put':True , 'rpc':True, 'uncached':False, 'audit': True},
                                 {'put':True , 'rpc':True, 'uncached':False, 'audit': True},
                                 {'put':False , 'rpc':False, 'uncached':False, 'audit': False},
                                 {'put':False , 'rpc':False, 'uncached':False, 'audit': False},
                                 ]):
            try:
                self.assertDictEqual(ch.perm, perm)
            except AssertionError as e:
                raise AssertionError('%s : %s'%(perm ,e))

class TestRE(unittest.TestCase):
    def test_offset(self):
        self.assertEqual(_sub_add(r'test', 1, 5), r'test')
        self.assertEqual(_sub_add(r'tes\\t', 1, 5), r'tes\\t')
        self.assertEqual(_sub_add(r'tes\1t', 1, 5), r'tes\6t')
        self.assertEqual(_sub_add(r'tes\1t\2', 2, 5), r'tes\6t\7')




from __future__ import print_function

import logging
import warnings
import sys
_log = logging.getLogger(__name__)

try:
    from itertools import izip
except ImportError:
    izip = zip
from functools import partial
import json
import threading

try:
    from Queue import Queue, Full, Empty
except ImportError:
    from queue import Queue, Full, Empty

from . import raw
from .raw import Disconnected, RemoteError, Cancelled, Finished, LazyRepr
from ..util import _defaultWorkQueue
from ..wrapper import Value, Type
from ..rpc import WorkQueue
from .._p4p import (logLevelAll, logLevelTrace, logLevelDebug,
                    logLevelInfo, logLevelWarn, logLevelError,
                    logLevelFatal, logLevelOff)

__all__ = [
    'Context',
    'Value',
    'Type',
    'RemoteError',
    'TimeoutError',
]

if sys.version_info >= (3, 0):
    unicode = str
    TimeoutError = TimeoutError

else:
    class TimeoutError(RuntimeError):
        "Local timeout has expired"
        def __init__(self):
            RuntimeError.__init__(self, 'Timeout')


class Subscription(object):
    """An active subscription.

    Returned by `Context.monitor`.
    """

    def __init__(self, ctxt, name, cb, notify_disconnect=False, queue=None):
        self.name, self._S, self._cb = name, None, cb
        self._notify_disconnect = notify_disconnect
        self._Q = queue or ctxt._Q or _defaultWorkQueue()
        self._evt = threading.Event()
        if notify_disconnect:
            # all subscriptions are inittially disconnected
            self._Q.push_wait(partial(cb, Disconnected()))

    def close(self):
        """Close subscription.
        """
        if self._S is not None:
            # after .close() self._event should never be called
            self._S.close()
            # wait for Cancelled to be delivered
            self._evt.wait()
            self._S = None

    def __enter__(self):
        return self

    def __exit__(self, A, B, C):
        self.close()

    @property
    def done(self):
        'Has all data for this subscription been received?'
        return self._S is None or self._S.done()

    @property
    def empty(self):
        'Is data pending in event queue?'
        return self._S is None or self._S.empty()

    def _event(self, E):
        try:
            assert self._S is not None, self._S
            # TODO: ensure ordering of error and data events
            _log.debug('Subscription wakeup for %s with %s', self.name, LazyRepr(E))
            self._inprog = True
            self._Q.push(partial(self._handle, E))
        except:
            _log.exception("Lost Subscription update: %s", LazyRepr(E))

    def _handle(self, E):
        try:
            S = self._S

            if isinstance(E, Cancelled):
                self._evt.set()
                return

            elif isinstance(E, (Disconnected, RemoteError)):
                _log.debug('Subscription notify for %s with %s', self.name, E)
                if self._notify_disconnect:
                    self._cb(E)
                elif isinstance(E, RemoteError):
                    _log.error("Subscription Error %s", E)
                return

            elif S is None:  # already close()'d
                return

            for n in range(4):
                E = S.pop()
                if E is None:
                    break
                self._cb(E)

            if E is not None:
                # removed 4 elements without emptying queue
                # re-schedule to mux with others
                self._Q.push(partial(self._handle, True))
            elif S.done:
                _log.debug('Subscription complete %s', self.name)
                S.close()
                S = None
                if self._notify_disconnect:
                    self._cb(Finished())
        except:
            _log.exception("Error processing Subscription event: %s", LazyRepr(E))
            if self._S is not None:
                self._S.close()
            self._S = None


class Context(raw.Context):

    """Context(provider, conf=None, useenv=True)

    :param str provider: A Provider name.  Try "pva" or run :py:meth:`Context.providers` for a complete list.
    :param dict conf: Configuration to pass to provider.  Depends on provider selected.
    :param bool useenv: Allow the provider to use configuration from the process environment.
    :param int workers: Size of thread pool in which monitor callbacks are run.  Default is 4
    :param int maxsize: Size of internal work queue used for monitor callbacks.  Default is unlimited
    :param dict nt: Controls :ref:`unwrap`.  None uses defaults.  Set False to disable
    :param dict unwrap: Legacy :ref:`unwrap`.
    :param WorkQueue queue: A work queue through which monitor callbacks are dispatched.

    The methods of this Context will block the calling thread until completion or timeout

    The meaning, and allowed keys, of the configuration dictionary depend on the provider.
    conf= will override values taken from the process environment.  Pass useenv=False to
    ensure that environment variables are completely ignored.

    The "pva" provider understands the following keys:

    * EPICS_PVA_ADDR_LIST
    * EPICS_PVA_AUTO_ADDR_LIST
    * EPICS_PVA_SERVER_PORT
    * EPICS_PVA_BROADCAST_PORT
    """
    Value = Value

    name = ''
    "Provider name string"

    def __init__(self, provider='pva', conf=None, useenv=True, nt=None, unwrap=None,
                 maxsize=0, queue=None):
        self._channel_lock = threading.Lock()

        super(Context, self).__init__(provider, conf=conf, useenv=useenv, nt=nt, unwrap=unwrap)

        # lazy start threaded WorkQueue
        self._Q = self._T = None

        self._Q = queue

    def _channel(self, name):
        with self._channel_lock:
            return super(Context, self)._channel(name)

    def disconnect(self, *args, **kws):
        with self._channel_lock:
            super(Context, self).disconnect(*args, **kws)

    def _queue(self):
        if self._Q is None:
            Q = WorkQueue(maxsize=self._Qmax)
            Ts = []
            for n in range(self._Wcnt):
                T = threading.Thread(name='p4p Context worker', target=Q.handle)
                T.daemon = True
                Ts.append(T)
            for T in Ts:
                T.start()
            _log.debug('Started %d Context worker', self._Wcnt)
            self._Q, self._T = Q, Ts
        return self._Q

    def close(self):
        """Force close all Channels and cancel all Operations
        """
        if self._Q is not None:
            for T in self._T:
                self._Q.interrupt()
            for n, T in enumerate(self._T):
                _log.debug('Join Context worker %d', n)
                T.join()
            _log.debug('Joined Context workers')
            self._Q, self._T = None, None
        if not Context:
            # Python 2.7 GC removes Context from scope during destruction of objects.
            return
        super(Context, self).close()

    def get(self, name, request=None, timeout=5.0, throw=True):
        """Fetch current value of some number of PVs.

        :param name: A single name string or list of name strings
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param float timeout: Operation timeout in seconds
        :param bool throw: When true, operation error throws an exception.  If False then the Exception is returned instead of the Value

        :returns: A p4p.Value or Exception, or list of same.  Subject to :py:ref:`unwrap`.

        When invoked with a single name then returns is a single value.
        When invoked with a list of name, then returns a list of values

        >>> ctxt = Context('pva')
        >>> V = ctxt.get('pv:name')
        >>> A, B = ctxt.get(['pv:1', 'pv:2'])
        >>>
        """
        singlepv = isinstance(name, (bytes, unicode))
        if singlepv:
            name = [name]
            request = [request]

        elif request is None:
            request = [None] * len(name)

        assert len(name) == len(request), (name, request)

        # use Queue instead of Event to allow KeyboardInterrupt
        done = Queue()
        result = [TimeoutError()] * len(name)
        ops = [None] * len(name)

        raw_get = super(Context, self).get

        try:
            for i, (N, req) in enumerate(izip(name, request)):
                def cb(value, i=i):
                    try:
                        if not isinstance(value, Cancelled):
                            done.put_nowait((value, i))
                        _log.debug('get %s Q %s', N, LazyRepr(value))
                    except:
                        _log.exception("Error queuing get result %s", value)

                _log.debug('get %s w/ %s', N, req)
                ops[i] = raw_get(N, cb, request=req)

            for _n in range(len(name)):
                try:
                    value, i = done.get(timeout=timeout)
                except Empty:
                    if throw:
                        _log.debug('timeout %s after %s', name[i], timeout)
                        raise TimeoutError()
                    break
                _log.debug('got %s %s', name[i], LazyRepr(value))
                if throw and isinstance(value, Exception):
                    raise value
                result[i] = value

        finally:
            [op and op.close() for op in ops]

        if singlepv:
            return result[0]
        else:
            return result

    def put(self, name, values, request=None, timeout=5.0, throw=True,
            process=None, wait=None, get=True):
        """Write a new value of some number of PVs.

        :param name: A single name string or list of name strings
        :param values: A single value, a list of values, a dict, a `Value`.  May be modified by the constructor nt= argument.
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param float timeout: Operation timeout in seconds
        :param bool throw: When true, operation error throws an exception.
                     If False then the Exception is returned instead of the Value
        :param str process: Control remote processing.  May be 'true', 'false', 'passive', or None.
        :param bool wait: Wait for all server processing to complete.
        :param bool get: Whether to do a Get before the Put.  If True then the value passed to the builder callable
                         will be initialized with recent PV values.  eg. use this with NTEnum to find the enumeration list.

        :returns: A None or Exception, or list of same

        When invoked with a single name then returns is a single value.
        When invoked with a list of name, then returns a list of values

        If 'wait' or 'process' is specified, then 'request' must be omitted or None.

        >>> ctxt = Context('pva')
        >>> ctxt.put('pv:name', 5.0)
        >>> ctxt.put(['pv:1', 'pv:2'], [1.0, 2.0])
        >>> ctxt.put('pv:name', {'value':5})
        >>>

        The provided value(s) will be automatically coerced to the target type.
        If this is not possible then an Exception is raised/returned.

        Unless the provided value is a dict, it is assumed to be a plain value
        and an attempt is made to store it in '.value' field.
        """
        if request and (process or wait is not None):
            raise ValueError("request= is mutually exclusive to process= or wait=")
        elif process or wait is not None:
            request = 'field()record[block=%s,process=%s]' % ('true' if wait else 'false', process or 'passive')

        singlepv = isinstance(name, (bytes, unicode))
        if singlepv:
            name = [name]
            values = [values]
            request = [request]

        elif request is None:
            request = [None] * len(name)

        assert len(name) == len(request), (name, request)
        assert len(name) == len(values), (name, values)

        # use Queue instead of Event to allow KeyboardInterrupt
        done = Queue()
        result = [TimeoutError()] * len(name)
        ops = [None] * len(name)

        raw_put = super(Context, self).put

        try:
            for i, (n, value, req) in enumerate(izip(name, values, request)):
                if isinstance(value, (bytes, unicode)) and value[:1] == '{':
                    try:
                        value = json.loads(value)
                    except ValueError:
                        raise ValueError("Unable to interpret '%s' as json" % value)

                # completion callback
                def cb(value, i=i):
                    try:
                        done.put_nowait((value, i))
                    except:
                        _log.exception("Error queuing put result %s", LazyRepr(value))

                ops[i] = raw_put(n, cb, builder=value, request=req, get=get)

            for _n in range(len(name)):
                try:
                    value, i = done.get(timeout=timeout)
                except Empty:
                    if throw:
                        raise TimeoutError()
                    break
                if throw and isinstance(value, Exception):
                    raise value
                result[i] = value

            if singlepv:
                return result[0]
            else:
                return result
        finally:
            [op and op.close() for op in ops]

    def rpc(self, name, value, request=None, timeout=5.0, throw=True):
        """Perform a Remote Procedure Call (RPC) operation

        :param str name: PV name string
        :param Value value: Arguments.  Must be Value instance
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param float timeout: Operation timeout in seconds
        :param bool throw: When true, operation error throws an exception.
                     If False then the Exception is returned instead of the Value

        :returns: A Value or Exception.  Subject to :py:ref:`unwrap`.

        >>> ctxt = Context('pva')
        >>> ctxt.rpc('pv:name:add', {'A':5, 'B'; 6})
        >>>

        The provided value(s) will be automatically coerced to the target type.
        If this is not possible then an Exception is raised/returned.

        Unless the provided value is a dict, it is assumed to be a plain value
        and an attempt is made to store it in '.value' field.
        """
        done = Queue()

        op = super(Context, self).rpc(name, done.put_nowait, value, request=request)

        try:
            try:
                result = done.get(timeout=timeout)
            except Empty:
                result = TimeoutError()
            if throw and isinstance(result, Exception):
                raise result

            return result
        except:
            op.close()
            raise

    def monitor(self, name, cb, request=None, notify_disconnect=False, queue=None):
        """Create a subscription.

        :param str name: PV name string
        :param callable cb: Processing callback
        :param request: A :py:class:`p4p.Value` or string to qualify this request, or None to use a default.
        :param bool notify_disconnect: In additional to Values, the callback may also be call with instances of Exception.
                                       Specifically: Disconnected , RemoteError, or Cancelled
        :param WorkQueue queue: A work queue through which monitor callbacks are dispatched.
        :returns: a :py:class:`Subscription` instance

        The callable will be invoked with one argument which is either.

        * A p4p.Value (Subject to :py:ref:`unwrap`)
        * A sub-class of Exception (Disconnected , RemoteError, or Cancelled)
        """
        R = Subscription(self, name, cb, notify_disconnect=notify_disconnect, queue=queue)

        R._S = super(Context, self).monitor(name, R._event, request)
        return R




from __future__ import print_function

import logging
_log = logging.getLogger(__name__)

import sys
import time
import json
try:
    from itertools import izip
except ImportError:
    izip = zip

import logging
_log = logging.getLogger(__name__)

from .. import Value
from .. import nt
from .. import set_debug
from . import thread


def op_get(ctxt, args):
    requests = [args.request] * len(args.names)
    results = ctxt.get(args.names, requests, timeout=args.timeout, throw=False)
    ret = 0
    for name, val in izip(args.names, results):
        if isinstance(val, Exception):
            ret = 1
            print(name, 'Error:', val)
        else:
            print(name, val)
    sys.exit(ret)


def op_put(ctxt, args):
    requests = [args.request] * len(args.names)

    names, values = [], []
    for pair in args.names:
        N, sep, V = pair.partition('=')
        if sep is '':
            print("Missing expected '=' after", pair)
            sys.exit(1)
        elif V[:1] in '{[':
            V = json.loads(V)
        N = N.strip()
        _log.debug("put %s <- %s", N, V)
        names.append(N)
        values.append(V)

    results = ctxt.put(names, values, requests, timeout=args.timeout, throw=False)

    ret = 0
    for name, val in izip(args.names, results):
        if isinstance(val, Exception):
            ret = 1
            print(name, 'Error:', val)
        elif val is None:
            print(name, 'ok')
        else:
            print(name, val.tolist())
    sys.exit(ret)


def op_monitor(ctxt, args):
    subs = []
    ret = 0

    for name in args.names:
        def show(val, name=name):
            if val is None:
                print(name, "Disconnect")
            elif isinstance(val, Exception):
                ret = 1
                print(name, 'Error:', val)
            else:
                print(name, val)
        subs.append(ctxt.monitor(name, show, args.request, notify_disconnect=True))

    try:
        while True:
            time.sleep(10)
    except KeyboardInterrupt:
        ret = 1
    [S.close() for S in subs]
    sys.exit(ret)


def op_rpc(ctxt, args):
    anames = []
    kws = {}
    for arg in args.args:
        K, sep, V = arg.partition('=')
        if not sep:
            print("arguments must be name=value not:", arg)
            sys.exit(2)
        elif V[:1] in '{[':
            V = json.loads(V)

        anames.append((K, 's'))
        kws[K] = V

    uri = nt.NTURI(anames).wrap(args.name, kws=kws)  # only keyword arguments

    ret = ctxt.rpc(args.name, uri, request=args.request, timeout=args.timeout, throw=False)
    if isinstance(ret, Exception):
        print('Error:', ret)
        sys.exit(1)
    else:
        print(ret.tolist())


def getargs():
    from argparse import ArgumentParser
    P = ArgumentParser()
    P.add_argument('-r', '--request', default='')
    P.add_argument('-w', '--timeout', type=float, default=5.0)
    P.add_argument('-p', '--provider', default='pva')
    P.add_argument('-d', '--debug', action='store_const', const=logging.DEBUG, default=logging.INFO)
    P.add_argument('-v', '--verbose', action='store_const', const=logging.DEBUG, default=logging.INFO)
    P.add_argument('--raw', action='store_false', default=None)
    P.set_defaults(func=lambda ctxt, args: P.print_help())

    SP = P.add_subparsers()

    PP = SP.add_parser('get')
    PP.add_argument('names', nargs='*')
    PP.set_defaults(func=op_get)

    PP = SP.add_parser('put')
    PP.add_argument('names', nargs='*', metavar='name=value', help='PV names and values')
    PP.set_defaults(func=op_put)

    PP = SP.add_parser('monitor')
    PP.add_argument('names', nargs='*')
    PP.set_defaults(func=op_monitor)

    PP = SP.add_parser('rpc')
    PP.add_argument('name')
    PP.add_argument('args', nargs='*')
    PP.set_defaults(func=op_rpc)

    return P.parse_args()


def main(args):
    with thread.Context(args.provider, unwrap=args.raw) as ctxt:
        args.func(ctxt, args)

if __name__ == '__main__':
    args = getargs()
    set_debug(args.debug)
    logging.basicConfig(level=args.verbose)
    main(args)



# distutils: language = c++
#cython: language_level=2

cimport cython

from libc.stddef cimport size_t
from libc.string cimport strchr
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.list cimport list as listxx
from libcpp.set cimport set as setxx

from cpython.object cimport PyObject, PyTypeObject, traverseproc, visitproc
from cpython.ref cimport Py_INCREF, Py_XDECREF

cdef extern from "<epicsAtomic.h>":
    cdef void atomic_set "::epics::atomic::set" (int& var, int val)

## PVD explicitly uses std::tr1::shared_ptr<> which is _sometimes_ a
## distinct type, and sometimes an alias for std::shared_ptr<>.
## So instead of:
#from libcpp.memory cimport shared_ptr, weak_ptr
## We replicate a minimal (noexcept) subset of memory.pxd for std::tr1::shared_ptr
cdef extern from "<pv/sharedPtr.h>" namespace "std::tr1" nogil:
    cdef cppclass shared_ptr[T]:
        shared_ptr()
        void reset()
        T* get()
        void swap(shared_ptr&)
        long use_count()
        bool unique()
        # not c++98...
        #bool operator bool()
        bool operator!()
    cdef cppclass weak_ptr[T]:
        weak_ptr()
        bool expired()
        shared_ptr[T] lock()

cdef extern from "<pv/security.h>" namespace "epics::pvAccess" nogil:
    cdef cppclass PeerInfo:
        string peer
        string transport
        string authority
        string realm
        string account
        setxx[string] roles
        unsigned transportVersion
        bool local
        bool identified

cdef extern from "<pv/pvAccess.h>" namespace "epics::pvAccess" nogil:
    cdef cppclass ChannelProvider:
        pass
    cdef cppclass ChannelRequester:
        shared_ptr[const PeerInfo] getPeerInfo() except+
    cdef cppclass ChannelProviderFactory:
        pass
    cdef cppclass ChannelProviderRegistry:
        @staticmethod
        shared_ptr[ChannelProviderRegistry] clients() except+
        shared_ptr[ChannelProviderFactory] remove(string& name) except+

cdef extern from "<pv/configuration.h>" namespace "epics::pvAccess" nogil:
    cdef cppclass Configuration:
        pass
    cdef cppclass ConfigurationBuilder:
        ConfigurationBuilder& add(const string& key, const string& value) except+
        ConfigurationBuilder& push_map() except+
        shared_ptr[Configuration] build() except+

cdef extern from "gwchannel.h" namespace "GWProvider" nogil:
    cdef struct ReportItem:
        string usname
        string dsname
        string transportPeer
        string transportAccount
        double transportTX
        double transportRX
        double operationTX
        double operationRX

cdef extern from "gwchannel.h" nogil:
    void GWInstallClientAliased(shared_ptr[ChannelProvider]& provider, string& installAs) except+

    cdef cppclass GWChan:

        int allow_put
        int allow_rpc
        int allow_uncached
        int audit
        int get_holdoff

        void disconnect()

    cdef struct GWStats:
        size_t ccacheSize
        size_t mcacheSize
        size_t gcacheSize
        size_t banHostSize
        size_t banPVSize
        size_t banHostPVSize

    enum: GWSearchIgnore
    enum: GWSearchClaim
    enum: GWSearchBanHost
    enum: GWSearchBanPV
    enum: GWSearchBanHostPV

    cdef cppclass GWProvider(ChannelProvider):
        PyObject* handle

        @staticmethod
        shared_ptr[ChannelProvider] buildClient(const string& name, const shared_ptr[Configuration]& conf) except+
        @staticmethod
        shared_ptr[GWProvider] build(const string& name, const shared_ptr[ChannelProvider]& provider) except+

        shared_ptr[GWProvider] shared_from_this() except+

        int test(const string& usname)
        shared_ptr[GWChan] connect(const string &dsname, const string &usname, const shared_ptr[ChannelRequester]& requester) except+

        void sweep() except+
        void disconnect(const string& usname) except+
        void forceBan(const string& host, const string& usname) except+
        void clearBan() except+
        void cachePeek(setxx[string]& names) except+
        void stats(GWStats& stats)
        void report(vector[ReportItem]& us, vector[ReportItem]& ds, double& period) except+

        @staticmethod
        void prepare() except+

GWProvider.prepare()

cdef class ClientInstaller(object):
    cdef string name
    cdef weak_ptr[ChannelProvider] provider

    def __enter__(self):
        cdef shared_ptr[ChannelProvider] prov = self.provider.lock()
        with nogil:
            if <bool>prov:
                GWInstallClientAliased(prov, self.name)
            prov.reset()

    def __exit__(self,A,B,C):
        with nogil:
            ChannelProviderRegistry.clients().get().remove(self.name)

cdef class Client(object):
    """Client(provider, config)
    GW Client.  Wraps a C++ class ChannelProvider.  Pass to `Provider` ctor.

    ``config`` dict takes the same keys as `p4p.client.thread.Context` with the "pva" provider.

    :param unicode provider: Client provider name from global register
    :param dict config: Configuration for new client instance.
    """
    cdef shared_ptr[ChannelProvider] provider

    def __init__(self, unicode provider, dict config):
        cdef ConfigurationBuilder builder
        cdef string name = provider.encode('utf-8')

        for K, V in config.items():
            builder.add(K.encode('utf-8'), V.encode('utf-8'))

        with nogil:
            self.provider = GWProvider.buildClient(name, builder.push_map().build())

    def __dealloc__(self):
        with nogil:
            self.provider.reset()

    def installAs(self, unicode name):
        """Install in process-wide client registry under an alias (eg. not "pva")
        """
        cdef ClientInstaller inst = ClientInstaller()
        inst.name = name.encode('utf-8')
        inst.provider = <weak_ptr[ChannelProvider]>self.provider
        return inst

cdef class InfoBase(object):
    cdef shared_ptr[const PeerInfo] info

    @property
    def peer(self):
        if <bool>self.info:
            return self.info.get().peer.decode('UTF-8')
        else:
            return u''

    @property
    def identified(self):
        if <bool>self.info:
            return self.info.get().identified
        else:
            return False

    @property
    def account(self):
        if <bool>self.info:
            return self.info.get().account.decode('UTF-8')
        else:
            return u''

    @property
    def roles(self):
        cdef list ret = []
        if <bool>self.info:
            for role in self.info.get().roles:
                ret.append(role.decode('UTF-8'))
        return ret

    def __dealloc__(self):
        with nogil:
            self.info.reset()

cdef class CreateOp(InfoBase):
    """Handle for in-progress Channel creation request
    """
    cdef readonly bytes name
    cdef weak_ptr[ChannelRequester] requester
    cdef weak_ptr[GWProvider] provider
    cdef object __weakref__

    def create(self, bytes name=None):
        """Create a Channel with a given upstream (server-side) name

        :param bytes name: Upstream name to use.  This is what the GW Client will search for.
        :returns: A `Channel`
        """
        cdef shared_ptr[GWChan] gwchan
        cdef Channel chan = Channel()
        cdef string dsname = self.name
        cdef string usname = name or self.name
        cdef shared_ptr[ChannelRequester] requester = self.requester.lock()
        cdef shared_ptr[GWProvider] provider = self.provider.lock()

        if <bool>requester and <bool>provider:
            with nogil:
                gwchan = provider.get().connect(dsname, usname, requester)

            if not gwchan:
                raise RuntimeError("GW Provider will not create %s -> %s"%(usname, dsname))

            chan.channel = <weak_ptr[GWChan]>gwchan
            chan.info = self.info
            return chan
        else:
            raise RuntimeError("Dead CreateOp")


cdef class Channel(InfoBase):
    """Wraps C++ class GWChan

    cf. `CreateOp.create()`
    """
    cdef readonly bytes name
    cdef weak_ptr[GWChan] channel
    cdef object __weakref__

    @property
    def expired(self):
        """Has this Channel become unused/disconnected?
        """
        return self.channel.expired()

    def access(self, put=None, rpc=None, uncached=None, audit=None, holdoff=None):
        """Configure access control permissions, and other restrictions, on this channel.

        :param put: None to leave unchanged.  bool to permit/deny PUT operations
        :param rpc: None to leave unchanged.  bool to permit/deny RPC operations
        :param uncached: None to leave unchanged.  bool to permit/deny cache bypass for GET/MONITOR
        :param audit: None to leave unchanged.  bool to enable/disable PUT logging
        :param holdoff: None to leave unchanged.  float value to set GET holdoff period
        """
        cdef shared_ptr[GWChan] ch = self.channel.lock()
        if not ch:
            return
        if put is not None:
            atomic_set(ch.get().allow_put, put==True)
        if rpc is not None:
            atomic_set(ch.get().allow_rpc, rpc==True)
        if uncached is not None:
            atomic_set(ch.get().allow_uncached, uncached==True)
        if audit:
            atomic_set(ch.get().audit, audit==True)
        if holdoff:
            atomic_set(ch.get().get_holdoff, holdoff*1000)

    def close(self):
        """Force disconnect this Channel
        """
        cdef shared_ptr[GWChan] ch = self.channel.lock()
        if <bool>ch:
            ch.get().disconnect()

@cython.no_gc_clear
cdef class Provider:
    """Provider(name, client, handler)
    GW Server endpoint.  wrapper for C++ class GWProvider

    :param unicode name: Unique name of this provider
    :param Client client: Associated client to which requests are forwarded
    :param handler: Callbacks
    """
    cdef shared_ptr[GWProvider] provider
    cdef object __weakref__
    cdef object dummy # ensure that this type participates in GC

    cdef readonly int Claim
    cdef readonly int Ignore
    cdef readonly int BanHost
    cdef readonly int BanPV
    cdef readonly int BanHostPV

    def __cinit__(self):
        self.Claim = GWSearchClaim
        self.Ignore = GWSearchIgnore
        self.BanHost = GWSearchBanHost
        self.BanPV = GWSearchBanPV
        self.BanHostPV = GWSearchBanHostPV

    def __init__(self, unicode name, Client client, object handler):
        cdef string cname = name.encode('utf-8')
        cdef shared_ptr[ChannelProvider] cprov

        cprov = client.provider
        with nogil:
            self.provider = GWProvider.build(cname, cprov)

        Py_INCREF(handler)
        Py_XDECREF(self.provider.get().handle)
        self.provider.get().handle = <PyObject*>handler

    def __dealloc__(self):
        with nogil:
            self.provider.reset()

    def testChannel(self, bytes usname):
        """testChannel(usname)
        Add the upstream name to the channel cache and begin trying to connect.
        Returns Claim if the channel is connected, and Ignore if it is not.

        :param bytes usname: Upstream (Server side) PV name
        :returns: Claim or Ignore
        """
        cdef string n = usname
        cdef int ret
        with nogil:
            ret = self.provider.get().test(n)
        return ret

    def sweep(self):
        """Call periodically to remove unused `Channel` from channel cache.
        """
        with nogil:
            self.provider.get().sweep()

    def disconnect(self, bytes usname):
        """Force disconnection of all channels connected to the named PV

        :param bytes usname: Upstream (Server side) PV name
        """
        cdef string n = usname
        with nogil:
            self.provider.get().disconnect(n)

    def forceBan(self, bytes host = None, bytes usname = None):
        """Preemptively Add an entry to the negative result cache.
        Either host or usname must be not None

        :param bytes host: None or a host name
        :param bytes usname: None or a upstream (Server side) PV name
        """
        cdef string h
        cdef string us
        if host:
            h = host
        if usname:
            us = usname
        with nogil:
            self.provider.get().forceBan(h, us)

    def clearBan(self):
        """Clear the negative results cache
        """
        with nogil:
            self.provider.get().clearBan()

    def cachePeek(self):
        """Returns PV names in channel cache

        :returns: a set of strings
        """
        cdef setxx[string] pvs
        cdef set ret

        self.provider.get().cachePeek(pvs)
        ret = set()
        for name in pvs:
            ret.add(name)
        return ret

    def stats(self):
        """Return statistics of various internal caches

        :rtype: dict
        """
        cdef GWStats stats
        self.provider.get().stats(stats)

        # cf. statsType in gw.py
        return {
            'ccacheSize.value':stats.ccacheSize,
            'mcacheSize.value':stats.mcacheSize,
            'gcacheSize.value':stats.gcacheSize,
            'banHostSize.value':stats.banHostSize,
            'banPVSize.value':stats.banPVSize,
            'banHostPVSize.value':stats.banHostPVSize,
        }

    def report(self):
        """Run bandwidth usage report

        :returns: A tuple of upstream (server side), downstream (client side), and time since last report() call.
        :rtype: ([(usname, opTx, opRx, peer, trTx, trRx)], [(usname, dsname, opTx, opRx, account, peer, trTx, trRx)], float)
        """
        cdef vector[ReportItem] us
        cdef vector[ReportItem] ds
        cdef double period = 0.0 # initial value not used, but quiets warning
        cdef ReportItem item

        with nogil:
            self.provider.get().report(us, ds, period)

        rus = []
        for item in us:
            # order in tuple must match column order
            rus.append((
                item.usname.decode('UTF-8'),
                item.operationTX,
                item.operationRX,
                item.transportPeer.decode('UTF-8'),
                item.transportTX,
                item.transportRX,
            ))

        rds = []
        for item in ds:
            # order in tuple must match column order
            rds.append((
                item.usname.decode('UTF-8'),
                item.dsname.decode('UTF-8'),
                item.operationTX,
                item.operationRX,
                item.transportAccount.decode('UTF-8'),
                item.transportPeer.decode('UTF-8'),
                item.transportTX,
                item.transportRX,
            ))

        return rus, rds, period

    def use_count(self):
        return self.provider.use_count()

# Allow GC to find handler stored in GWProvider
#   https://github.com/cython/cython/issues/2737
cdef traverseproc Provider_base_traverse

cdef int holder_traverse(PyObject* raw, visitproc visit, void* arg) except -1:
    cdef int ret = 0
    cdef Provider self = <Provider>raw

    if self.provider.get().handle:
        visit(self.provider.get().handle, arg)
    return Provider_base_traverse(raw, visit, arg)

Provider_base_traverse = (<PyTypeObject*>Provider).tp_traverse
(<PyTypeObject*>Provider).tp_traverse = <traverseproc>holder_traverse

# called from gwchannel.cpp
cdef public:
    void GWProvider_cleanup(GWProvider* provider) with gil:
        Py_XDECREF(provider.handle)

    int GWProvider_testChannel(GWProvider* provider, const char* name, const char* peer) with gil:
        if not provider.handle:
            return GWSearchBanHost
        handle = <object>provider.handle
        try:
            return handle.testChannel(name, peer.decode('UTF-8'))
        except:
            import traceback
            traceback.print_exc()
            return GWSearchBanHost

    shared_ptr[GWChan] GWProvider_makeChannel(GWProvider* provider, const string& name, const shared_ptr[ChannelRequester]& requester) with gil:
        cdef shared_ptr[GWChan] ret
        cdef CreateOp op
        cdef Channel result

        if provider.handle:
            op = CreateOp()
            handle = <object>provider.handle
            op.name = name.c_str()
            op.requester = <weak_ptr[ChannelRequester]>requester
            op.provider = <weak_ptr[GWProvider]>provider.shared_from_this()
            op.info = requester.get().getPeerInfo()

            try:
                chan = handle.makeChannel(op)
            except:
                import traceback
                traceback.print_exc()
                chan = None

            if chan is not None:
                ret = (<Channel?>chan).channel.lock()

        return ret

    void GWProvider_audit(GWProvider* provider, listxx[string]& audits) with gil:
        if provider.handle:
            handle = <object>provider.handle

            for audit in audits:
                try:
                    handle.audit(audit)
                except:
                    import traceback
                    traceback.print_exc()
                    break



# stripped down version of libcpp/set.pxd
# only supports const iterator (which is the only sane way to iterate a set)
cdef extern from "<set>" namespace "std" nogil:
    cdef cppclass set[T]:
        ctypedef T value_type
        cppclass const_iterator:
            T& operator*()
            const_iterator operator++()
            const_iterator operator--()
            bint operator==(const_iterator)
            bint operator!=(const_iterator)
        set() except +
        set(set&) except +
        const_iterator begin()
        void clear()
        size_t count(const T&)
        bint empty()
        const_iterator end()
        const_iterator find(T&)
        size_t size()
        void swap(set&)




from __future__ import print_function

import unittest
import random
import weakref
import sys
import gc
import threading

from ..wrapper import Value, Type
from ..client.thread import Context
from ..server import Server, installProvider, removeProvider
from ..rpc import NTURIDispatcher, WorkQueue, rpc, rpccall, rpcproxy
from ..nt import NTScalar, NTURI
from .utils import RefTestCase


class TestService(object):

    @rpc(NTScalar('d'))
    def add(self, lhs, rhs):
        return float(lhs) + float(rhs)


class TestRPCFull(RefTestCase):

    """Test end to end

    full server and client communicating through the loopback
    """
    runserver = True
    provider = 'pva'
    getconfig = lambda self: self.server.conf()

    def setUp(self):
        super(TestRPCFull, self).setUp()

        conf = {
            'EPICS_PVAS_INTF_ADDR_LIST': '127.0.0.1',
            'EPICS_PVA_ADDR_LIST': '127.0.0.1',
            'EPICS_PVA_AUTO_ADDR_LIST': '0',
            'EPICS_PVA_SERVER_PORT': '0',
            'EPICS_PVA_BROADCAST_PORT': '0',
        }

        # random PV prefix
        self.prefix = 'rpctest:%u:' % random.randint(0, 1024)

        service = TestService()

        # RPC handling queue
        self._Q = WorkQueue(maxsize=2)

        # RPC dispatcher (extract RPC args from PVD blob)
        dispatch = NTURIDispatcher(self._Q, target=service, prefix=self.prefix, name="TestRPC")
        self._dispatch = weakref.ref(dispatch)

        if self.runserver:
            self.server = Server(providers=[dispatch], conf=conf, useenv=False)
            print("conf", self.server.conf())
        else:
            installProvider("TestRPC", dispatch)

        self._QT = threading.Thread(name="TestRPC Q", target=self._Q.handle)
        self._QT.start()

    def tearDown(self):
        if self.runserver:
            self.server.stop()
            self.server = None

        self._Q.interrupt()
        self._QT.join()

        if not self.runserver:
            removeProvider("TestRPC")
        gc.collect()
        D = self._dispatch()
        if D is not None:
            print("dispatcher lives! ", sys.getrefcount(D), " refs  referrers:")
            import inspect
            for R in gc.get_referrers(D):
                print(R)
        self.assertIsNone(D)
        super(TestRPCFull, self).tearDown()

    def testAdd(self):
        args = NTURI([
            ('lhs', 'd'),
            ('rhs', 'd'),
        ]).wrap(self.prefix + 'add', kws={
            'lhs': 1,
            'rhs': 1,
        }, scheme='pva')
        with Context(self.provider, useenv=False, conf=self.getconfig(), unwrap=False) as ctxt:
            self.assertEqual(ctxt.name, self.provider)
            sum = ctxt.rpc(self.prefix + 'add', args)
            self.assertEqual(sum.value, 2.0)

    def testAdd3(self):
        args = NTURI([
            ('lhs', 'd'),
            ('rhs', 'd'),
        ]).wrap(self.prefix + 'add', kws={
            'lhs': 1,
            'rhs': 2,
        }, scheme='pva')
        with Context(self.provider, useenv=False, conf=self.getconfig(), unwrap=False) as ctxt:
            sum = ctxt.rpc(self.prefix + 'add', args)
            self.assertEqual(sum.value, 3.0)


class TestRPCProvider(TestRPCFull):

    """end to end w/o network
    """
    runserver = False
    provider = 'server:TestRPC'
    getconfig = lambda self: {}


class TestProxy(RefTestCase):

    class MockContext(object):
        name = 'fake'

        def rpc(self, *args, **kws):
            return args, kws

    @rpcproxy
    class MyProxy(object):

        def __init__(self, myarg=5):
            self.myarg = myarg

        @rpccall('%sfoo')
        def bar(A='i', B='s'):
            pass

        @rpccall('%sbaz')
        def another(X='s', Y='i'):
            pass

    def setUp(self):
        super(TestProxy, self).setUp()
        ctxt = self.MockContext()
        self.proxy = self.MyProxy(myarg=3, context=ctxt, format='pv:')
        self.assertEqual(self.proxy.myarg, 3)
        self.assertIs(self.proxy.context, ctxt)

    def test_call1(self):
        args, kws = self.proxy.bar(4, 'one')

        print(args, kws)
        self.assertEqual(args[0], 'pv:foo')
        self.assertListEqual(args[1].tolist(), [
            ('scheme', u'fake'),
            ('authority', u''),
            ('path', u'pv:foo'),
            ('query', [('A', 4), ('B', u'one')])
        ])
        self.assertEqual(len(args), 2)
        self.assertDictEqual(kws, {'request': None, 'throw': True, 'timeout': 3.0})

    def test_call2(self):
        args, kws = self.proxy.bar(4, B='one')

        print(args, kws)
        self.assertEqual(args[0], 'pv:foo')
        self.assertListEqual(args[1].tolist(), [
            ('scheme', u'fake'),
            ('authority', u''),
            ('path', u'pv:foo'),
            ('query', [('A', 4), ('B', u'one')])
        ])
        self.assertEqual(len(args), 2)
        self.assertDictEqual(kws, {'request': None, 'throw': True, 'timeout': 3.0})

    def test_call3(self):
        args, kws = self.proxy.bar(4)

        print(args, kws)
        self.assertEqual(args[0], 'pv:foo')
        self.assertListEqual(args[1].tolist(), [
            ('scheme', u'fake'),
            ('authority', u''),
            ('path', u'pv:foo'),
            ('query', [('A', 4)])
        ])
        self.assertEqual(len(args), 2)
        self.assertDictEqual(kws, {'request': None, 'throw': True, 'timeout': 3.0})

    def test_call4(self):
        args, kws = self.proxy.another('one', Y=2)

        print(args, kws)
        self.assertEqual(args[0], 'pv:baz')
        self.assertListEqual(args[1].tolist(), [
            ('scheme', u'fake'),
            ('authority', u''),
            ('path', u'pv:baz'),
            ('query', [('X', u'one'), ('Y', 2)])
        ])
        self.assertEqual(len(args), 2)
        self.assertDictEqual(kws, {'request': None, 'throw': True, 'timeout': 3.0})



import logging
import warnings
import socket
import re

from threading import Lock
from collections import defaultdict
from functools import partial
from weakref import WeakKeyDictionary

from .yacc import parse, ACFError

from .. import Value
from ..client.thread import Context, LazyRepr, Disconnected

_log = logging.getLogger(__name__)

NONE = 0
READ = 1
PUT = 2
RPC = 4
UNCACHED = 8
WRITE = READ | PUT | RPC
actionmask = {
    'NONE': NONE,
    'READ':READ,
    'WRITE':WRITE,
    'PUT':PUT,
    'RPC':RPC,
    'UNCACHED':UNCACHED,
}

class Engine(object):
    '''Access Security File (ACF) parsing and evaluation engine

    >>> with open(fname, 'r') as F:
            acf = Engine(F.read())
    '''

    defaultACF = """
    ASG(DEFAULT) {
        RULE(1, WRITE)
        RULE(1, UNCACHED)
    }
    """
    def __init__(self, acf = None, ctxt = None):
        self._lock = Lock()
        # {Channel:(group, user, host, level)}
        self._anodes = WeakKeyDictionary()
        self._ctxt = ctxt
        self._asg = {}
        self._subscriptions = {}

        self.parse(acf or self.defaultACF)

    def report(self):
        A, B, C, D = [], [], [], []
        for asg,(rules,inputs) in self._asg.items():
            for var, val in inputs.items():
                A.append(asg)
                B.append(var)
                C.append(val or 0.0)
                D.append(val is not None)
        return {
            'value.asg':A,
            'value.var':B,
            'value.value':C,
            'value.connected':D,
        }

    def parse(self, acf):
        ast = parse(acf)

        # map user or host to set of groups
        uag = defaultdict(set)
        hag = defaultdict(set)
        uags, hags = set(), set()

        # {'name';([rules], {vars})}
        asg = {}
        # {'pvname':[('name','VAR')]}
        invars = defaultdict(list)

        for node in ast:
            if node[0]=='UAG':
                # ('UAG', name, [members...])
                uags.add(node[1])
                for member in node[2]:
                    uag[member].add(node[1])

            elif node[0]=='HAG':
                # ('HAG', name, [members...])
                hags.add(node[1])
                for member in node[2]:
                    hag[member].add(node[1])

            elif node[0]=='ASG':
                # ('ASG', name, [rules...])
                #   rule : ('INP', 'A', 'pv:name')
                #        | ('RULE', 1, 'WRITE', trap, None | [])

                rules, inputs = asg[node[1]] = [], {}
                for anode in node[2]:
                    if anode[0]=='RULE':
                        rule = []
                        for rnode in anode[4] or []:
                            if rnode[0] in ('UAG', 'HAG'):
                                # ('UAG', ['name'])
                                # ('HAG', ['name'])
                                rule.append((rnode[0], set(rnode[1])))

                            elif rnode[0]=='CALC':
                                # ('CALC', '<expr>')
                                for var in re.findall(r'[A-Z]', rnode[1]):
                                    inputs[var] = None

                                # cheating here by using python expression syntax instead of CALC.
                                try:
                                    expr = compile(rnode[1], '<acf>', 'eval')
                                    _log.debug('Compile %s', rnode[1])
                                except SyntaxError:
                                    _log.exception("Error in CALC expression")
                                    # default to false on error
                                    expr = compile('0', '<acf>', 'eval')

                                rule.append((rnode[0], rnode[1], expr))

                            else:
                                warnings.warn("Invalid RULE condition AST: %s"%(rnode,))

                        try:
                            mask = actionmask[anode[2]]
                        except KeyError:
                            _log.warn('Ignoring unknown permission "%s"', anode[2])
                            mask = 0 # grant no permissions
                        rules.append( (mask, anode[1], anode[3], rule) )

                    elif anode[0]=='INP':
                        # ('INP', 'A', 'pv:name')
                        invars[anode[2]].append((node[1], anode[1]))

                    else:
                        warnings.warn("Invalid Rule AST: %s"%(anode,))

            else:
                warnings.warn("Invalid AST: %s"%(node,))

        hag_addr = self._resolve_hag(hag)

        # prevent accidental insertions
        uag = dict(uag)
        hag = dict(hag)
        invars = dict(invars)

        # at this point, success is assumed.
        # aka. errors will not be clean

        if invars and self._ctxt is None:
            for pv, grps in invars.items():
                _log.warning('No Client to connect to ACF %s for %s', pv, grps)

        # cancel any active subscriptions
        [S.close() for S in self._subscriptions.values()]

        with self._lock:
            self._uag = uag
            self._hag = hag
            self._asg = asg
            self._asg_DEFAULT = asg.get('DEFAULT', [])
            self._hag_addr = hag_addr

        self._recompute()

        # create new subscriptions
        # which will trigger a lot of recomputes
        if self._ctxt is not None:
            [_log.debug('subscribing to %s for %s', pv, grps) for pv,grps in invars.items()]
            self._subscriptions = {pv: self._ctxt.monitor(pv, partial(self._var_update, grps), notify_disconnect=True) for pv,grps in invars.items()}
        else:
            self._subscriptions = {}

    def _var_update(self, grps, value):
        # clear old value first
        val = None
        if not isinstance(value, Disconnected):
            try:
                val = float(value or 0.0)
            except:
                _log.exception('INP%s unable to store %s', grps, LazyRepr(value))

        _log.debug('Update INP%s <- %s', grps, val)

        with self._lock:
            for asg, var in grps:
                _rules, inputs = self._asg[asg]
                inputs[var] = val

        if grps:
            self._recompute(only={asg for asg,var in grps})

    def _recompute(self, only=None):
        _log.debug("Recompute %s", only or "all")
        anodes, self._anodes = self._anodes, WeakKeyDictionary()

        for channel, (group, user, host, level) in anodes.items():
            if only is None or group in only:
                self.create(channel, group, user, host, level)
            else:
                self._anodes[channel] = (group, user, host, level)

    @staticmethod
    def _gethostbyname(host):
        try:
            return socket.gethostbyname(host)
        except socket.gaierror as e:
            _log.warn( "Ignore invalid hostname \"%s\" : %s", host, e )

    def _resolve_hag(self, _hag):
        hag_addr = defaultdict(set)

        for host, groups in _hag.items():
            ip = self._gethostbyname(host)
            if ip is not None:
                hag_addr[ip] |= groups

        return dict(hag_addr)

    def resolve_hag(self):
        # TODO: racy.  How to make atomic w/o waiting for DNS lookup with lock?
        _hag_addr = self._resolve_hag(self._hag)
        with self._lock:
            self._hag_addr = _hag_addr

        self._recompute()

    def create(self, channel, group, user, host, level, roles=[]):
        # Default to restrictive.  Used in case of error
        perm = 0
        _log.debug('(re)create %s, %s, %s, %s, %s', channel, group, user, host, level)

        with self._lock:

            uags = self._uag.get(user, set())
            for role in roles:
                uags |= self._uag.get('role/'+role, set())
            hags = self._hag_addr.get(host, set())
            rules, inputs = self._asg.get(group, self._asg_DEFAULT)

            trapit = False
            try:
                for mask, asl, trap, conds in rules:
                    accept = True
                    for cond in conds:
                        if cond[0]=='UAG':
                            accept = len(cond[1].intersection(uags))
                        elif cond[0]=='HAG':
                            accept = len(cond[1].intersection(hags))
                        elif cond[0]=='CALC':
                            try:
                                accept = float(eval(cond[2], {}, inputs) or 0.0) >= 0.5 # horray for legacy... I mean compatibility
                            except:
                                # this could be any of a number of exceptions
                                # which all add up to the same.  Invalid expression
                                accept = False
                                _log.exception('Error evaluating: %s with %s', cond[1], [(k,v,type(v)) for k,v in inputs.items()])
                            else:
                                _log.debug('Evaluate %s with %s -> %s', cond, [(k,v,type(v)) for k,v in inputs.items()], accept)
                        else:
                            warnings.warn("Invalid AST RULE: %s"%cond)
                            accept = False

                        if not accept:
                            break

                    if accept:
                        trapit |= trap
                        perm |= mask

            except:
                _log.exception("Error while calculating ASG for %s, %s, %s, %s, %s",
                            channel, group, user, host, level)

            put = perm & PUT
            rpc = perm & RPC
            uncached = perm & UNCACHED

            channel.access(put=bool(put), rpc=bool(rpc), uncached=bool(uncached), audit=trapit)

            self._anodes[channel] = (group, user, host, level)

    def _check_host(self, hag, user, host):
        groups = self._hag_addr.get(host) or set()
        return hag in groups

    def _check_user(self, uag, user, host):
        groups = self._uag.get(user) or set()
        return uag in groups




from ..wrapper import Type

# common sub-structs
timeStamp = Type(id='time_t', spec=[
    ('secondsPastEpoch', 'l'),
    ('nanoseconds', 'i'),
    ('userTag', 'i'),
])
alarm = Type(id='alarm_t', spec=[
    ('severity', 'i'),
    ('status', 'i'),
    ('message', 's'),
])



from __future__ import print_function

import logging
import sys
import unittest

from functools import partial
from collections import OrderedDict

from ..wrapper import Value, Type
from .. import nt
from .utils import RefTestCase

import numpy
from numpy.testing import assert_array_almost_equal as assert_aequal

_log = logging.getLogger(__name__)

class TestScalar(RefTestCase):

    def test_float_wrap(self, code='d', value=5.0):
        NT = nt.NTScalar(code)

        V = NT.wrap(value)
        self.assertEqual(V.value, value)
        self.assertEqual(V.alarm.severity, 0)
        self.assertIsNone(V.get('display'))

        NT = nt.NTScalar(code, display=True)
        V = NT.wrap({
            'value': value,
            'alarm': {
                'severity': 1,
            },
        })

        self.assertEqual(V.value, value)
        self.assertEqual(V.alarm.severity, 1)
        if code!='s':
            self.assertEqual(V.display.tolist(), [
                ('limitLow', 0.0),
                ('limitHigh', 0.0),
                ('description', u''),
                ('format', u''),
                ('units', u'')
            ])

    def test_int_wrap(self):
        self.test_float_wrap(code='i', value=42)
    def test_str_wrap(self):
        self.test_float_wrap(code='s', value='foo')

    def test_float_unwrap(self, code='d', value=5.0):
        NT = nt.NTScalar(code)
        V = NT.wrap({
            'value': value,
            'alarm': {
                'severity': 1,
            },
        })

        P = nt.NTScalar.unwrap(V)
        self.assertEqual(P, value)
        self.assertEqual(P.severity, 1)

        V2 = NT.wrap(P)
        self.assertIs(V, V2)

    def test_int_unwrap(self):
        self.test_float_unwrap(code='i', value=42)
    def test_str_unwrap(self):
        self.test_float_unwrap(code='s', value='foo')

    def test_array_wrap(self):
        NT = nt.NTScalar('ad')  # array of double

        A = numpy.asarray([1.0, 5.0])
        V = NT.wrap(A)
        assert_aequal(V.value, A)
        self.assertEqual(V.alarm.severity, 0)
        self.assertTrue(V.changed('value'))

    def test_array_unwrap(self):
        NT = nt.NTScalar('ad')  # array of double
        A = numpy.asarray(range(10))[2:5]
        V = NT.wrap(A)

        P = nt.NTScalar.unwrap(V)
        self.assertIsNot(P, A)
        assert_aequal(P, A)
        self.assertEqual(P.severity, 0)

    def test_string_array_wrap(self):
        NT = nt.NTScalar('as')  # array of string

        A = ["hello", "world"]
        V = NT.wrap(A)
        self.assertEqual(V.value, A)
        self.assertEqual(V.alarm.severity, 0)
        self.assertTrue(V.changed('value'))


class TestTable(RefTestCase):

    def test_wrap(self):
        NT = nt.NTTable(columns=[
            ('a', 'i'),
            ('b', 's'),
        ])
        V = NT.wrap([
            {'a': 5, 'b': 'one'},
            {'a': 6, 'b': 'two'},
        ])

        assert_aequal(V.value.a, [5, 6])
        self.assertEqual(V.value.b, ['one', 'two'])

    def test_unwrap(self):
        T = nt.NTTable.buildType(columns=[
            ('a', 'ai'),
            ('b', 'as'),
        ])
        V = Value(T, {
            'labels': ['a', 'b'],
            'value': {
                'a': [5, 6],
                'b': ['one', 'two'],
            },
        })

        P = list(nt.NTTable.unwrap(V))

        self.assertListEqual(P, [
            OrderedDict([('a', 5), ('b', u'one')]),
            OrderedDict([('a', 6), ('b', u'two')]),
        ])


class TestURI(RefTestCase):

    def test_build(self):
        NT = nt.NTURI([
            ('a', 'I'),
            ('b', 's'),
            ('c', ('S', None, [
                ('x', 'd'),
                ('y', 'd'),
            ])),
        ])

        V = NT.wrap('fn', (5,))
        self.assertEqual(V.query.a, 5)
        self.assertRaises(AttributeError, lambda: V.query.b)
        self.assertRaises(AttributeError, lambda: V.query.c)

        V = NT.wrap('fn', (6, 'foo'))
        self.assertEqual(V.query.a, 6)
        self.assertEqual(V.query.b, 'foo')
        self.assertRaises(AttributeError, lambda: V.query.c)

        V = NT.wrap('fn', (7,), {'b': 'bar'})
        self.assertEqual(V.query.a, 7)
        self.assertEqual(V.query.b, 'bar')
        self.assertRaises(AttributeError, lambda: V.query.c)

        V = NT.wrap('fn', (), {'a': 8, 'b': 'bar'})
        self.assertEqual(V.query.a, 8)
        self.assertEqual(V.query.b, 'bar')
        self.assertRaises(AttributeError, lambda: V.query.c)

        V = NT.wrap('fn', (), {'a': 8, 'b': 'bar', 'c': {'x': 1, 'y': 2}})
        self.assertEqual(V.query.a, 8)
        self.assertEqual(V.query.c.x, 1)
        self.assertEqual(V.query.c.y, 2)


class TestEnum(RefTestCase):

    def testStore(self):
        T = nt.NTEnum.buildType()

        V = Value(T, {
            'value.choices': ['zero', 'one', 'two'],
        })

        V.value = 'one'

        self.assertEqual(V.value.index, 1)

        V.value = '2'

        self.assertEqual(V.value.index, 2)

        V.value = 1

        self.assertEqual(V.value.index, 1)

    def testStoreBad(self):
        V = Value(nt.NTEnum.buildType(), {
            'value.choices': ['zero', 'one', 'two'],
        })

        V.value.index = 42

        def fn():
            V.value = self
        self.assertRaises(TypeError, fn)

        self.assertEqual(V.value.index, 42)

        def fn():
            V.value = 'other'
        self.assertRaises(ValueError, fn)

        self.assertEqual(V.value.index, 42)

        if sys.version_info >= (3, 0):
            V.value.choices = []

            def fn():
                V.value = '1'
            self.assertWarns(UserWarning, fn)  # warns of empty choices

            self.assertEqual(V.value.index, 1)

    def testSubStore(self):
        V = Value(Type([
            ('a', nt.NTEnum.buildType()),
            ('b', nt.NTEnum.buildType()),
        ]), {
            'a.value.choices': ['A', 'B'],
            'b.value.choices': ['X', 'Y'],
        })

        V.a = {'value': 'B'}

        self.assertEqual(V.a.value.index, 1)
        self.assertEqual(V.b.value.index, 0)

    def testWrap(self):
        W = nt.NTEnum()
        V = W.wrap({'index':1, 'choices':['X','Y']})

        self.assertEqual(V.value.index, 1)
        self.assertEqual(V.value.choices, ['X','Y'])

        W = nt.NTEnum()
        V = W.wrap(0)

        self.assertEqual(V.value.index, 0)
        self.assertEqual(V.value.choices, [])

    def testAssign(self):
        W = nt.NTEnum()
        V = nt.NTEnum.buildType()()
        W.assign(V, 1)

        self.assertEqual(V.value.index, 1)
        self.assertEqual(V.value.choices, [])

        V.value.choices = ['A', 'B']
        W.assign(V, 'A')

        self.assertEqual(V.value.index, 0)
        self.assertEqual(V.value.choices, ['A', 'B'])

    def testUnwrap(self):
        W = nt.NTEnum()
        V = nt.NTEnum.buildType()()

        U = W.unwrap(V)
        self.assertEqual(U, 0)
        self.assertIsNone(U.choice)

        V.value.index = 1
        V.value.choices = ['A', 'B']
        U = W.unwrap(V)
        self.assertEqual(U, 1)
        self.assertEqual(U.choice, 'B')
        self.assertEqual(W._choices, ['A', 'B'])

class TestArray(RefTestCase):

    def test_unwrap_None(self):
        V = Value(nt.NTNDArray.buildType(), {})

        img = nt.NTNDArray.unwrap(V)

        self.assertIsInstance(img, numpy.ndarray)
        assert_aequal(img.shape, (0,))

    def test_zero_length(self):
        V = Value(nt.NTNDArray.buildType(), {
            'value': numpy.arange(0),
            'dimension': [
                {'size': 3}, # X, columns
                {'size': 0}, # Y, rows
            ],
            'attribute': [
                {'name': 'ColorMode', 'value': 0},
            ],
        })

        img = nt.NTNDArray.unwrap(V)

        self.assertIsInstance(img, numpy.ndarray)
        assert_aequal(img.shape, (0,3))

    def test_unwrap_mono(self):
        pixels = numpy.asarray([  # 2x3
            [0, 1, 2],
            [3, 4, 5],
        ])
        # check my understanding of numpy
        self.assertTupleEqual(pixels.shape, (2, 3)) # inner-most right (in a pixel loop)
        assert_aequal(pixels.flatten(), [0, 1, 2, 3, 4, 5]) # row major

        V = Value(nt.NTNDArray.buildType(), {
            'value': numpy.arange(6),
            'dimension': [
                {'size': 3}, # X, columns
                {'size': 2}, # Y, rows
            ],
            'attribute': [
                {'name': 'ColorMode', 'value': 0},
            ],
        })

        img = nt.NTNDArray.unwrap(V)

        self.assertEqual(img.shape, (2, 3))
        assert_aequal(img, pixels)

        V2 = nt.NTNDArray().wrap(img)

        assert_aequal(V.value, V2.value)
        self.assertEqual(V.dimension[0].size, V2.dimension[0].size)
        self.assertEqual(V.dimension[1].size, V2.dimension[1].size)

    def test_unwrap_3d(self):
        # scipy.misc.face().shape==(768, 1024, 3)
        # in AD world this is [3, 1024, 768) w/ RGB1

        # we test with 4x2 RGB1
        pixels = numpy.array([
            [(1,0,0), (2,0,0), (3,0,0), (4,0,0)],
            [(5,0,0), (6,0,0), (7,0,0), (8,0,0)],
        ], dtype='u1')
        self.assertEqual(pixels.shape, (2, 4, 3))

        # manually construct matching NTNDArray
        V = Value(nt.NTNDArray.buildType(), {
            'value': ('ubyteValue', pixels.flatten()),
            'dimension': [
                {'size': 3}, # "color"
                {'size': 4}, # X, columns
                {'size': 2}, # Y, rows
            ],
            'attribute': [
                {'name': 'ColorMode', 'value': 2},
            ],
        })
        self.assertEqual(V.value.dtype, pixels.dtype)

        img = nt.NTNDArray.unwrap(V)

        self.assertEqual(img.shape, (2, 4, 3))
        assert_aequal(img, pixels)
        self.assertEqual(img.dtype, pixels.dtype)
        self.assertDictEqual(img.attrib, {u'ColorMode':2}) # RGB1

        # round trip
        V2 = nt.NTNDArray().wrap(img)

        assert_aequal(V.value, V2.value)
        self.assertEqual(V.dimension[0].size, V2.dimension[0].size)
        self.assertEqual(V.dimension[1].size, V2.dimension[1].size)
        self.assertEqual(V.dimension[2].size, V2.dimension[2].size)
        self.assertEqual(V2.value.dtype, pixels.dtype)
        self.assertEqual(V2.attribute[0].name, u'ColorMode')
        self.assertEqual(V2.attribute[0].value, 2)

        # wrap up raw array  (no pixels.attrib)
        V2 = nt.NTNDArray().wrap(pixels)

        assert_aequal(V.value, V2.value)
        self.assertEqual(V.dimension[0].size, V2.dimension[0].size)
        self.assertEqual(V.dimension[1].size, V2.dimension[1].size)
        self.assertEqual(V.dimension[2].size, V2.dimension[2].size)
        self.assertEqual(V2.value.dtype, pixels.dtype)
        self.assertEqual(V2.attribute[0].name, u'ColorMode')
        self.assertEqual(V2.attribute[0].value, 2)



import logging
import sys
import atexit

try:
    # give a chance to adjust DSO loader path
    import epicscorelibs.path
except ImportError:
    pass

from .wrapper import Value, Type
from ._p4p import (pvdVersion, pvaVersion, listRefs, Cancelled, ClientProvider as _ClientProvider)

from ._p4p import (logLevelAll, logLevelTrace, logLevelDebug,
                   logLevelInfo, logLevelWarn, logLevelError,
                   logLevelFatal, logLevelOff)

_log = logging.getLogger(__name__)

__all__ = (
    'Value',
    'Type',
)

_lvls = {
    logLevelAll, # 0
    logLevelTrace,
    logLevelDebug,
    logLevelInfo,
    logLevelWarn,
    logLevelError,
    logLevelFatal,
    logLevelOff, # 7
}

_lvlmap = {
    logging.DEBUG:logLevelDebug, # 10 -> 2
    logging.INFO:logLevelInfo,
    logging.WARN:logLevelWarn,
    logging.ERROR:logLevelError,
    logging.FATAL:logLevelFatal,
}

def set_debug(lvl):
    """Set PVA global debug print level.  This prints directly to stdout,
    bypassing eg. sys.stdout.

    :param lvl: logging.* level or logLevel*
    """
    lvl = _lvlmap.get(lvl, lvl)
    assert lvl in _lvls, lvl
    _ClientProvider.set_debug(lvl)

version = (1, 0, -80)

def cleanup():
    """P4P sequenced shutdown.  Intended to be atexit.  Idenpotent.
    """
    _log.debug("P4P atexit begins")
    # clean provider registry
    from .server import clearProviders, _cleanup_servers
    clearProviders()

    # close client contexts
    from .client.raw import _cleanup_contexts
    _cleanup_contexts()

    # stop servers
    _cleanup_servers()

    # shutdown default work queue
    from .util import _defaultWorkQueue
    _defaultWorkQueue.stop()
    _log.debug("P4P atexit completes")

atexit.register(cleanup)



#!/usr/bin/env python

from __future__ import print_function

import sysconfig

from setuptools_dso import Extension, setup, cythonize

import numpy
from numpy.distutils.misc_util import get_numpy_include_dirs

import epicscorelibs.path
import epicscorelibs.version
from epicscorelibs.config import get_config_var

# the following line is matched from cibuild.py
package_version = '3.5.4'

cxxflags = []
ldflags = []
import sys
import platform
if sys.platform=='linux2' and not sysconfig.get_config_var('Py_DEBUG'):
    # c++ debug symbols size is huge.  ~20x code size.
    # So we choose to only emit debug symbols when building for an interpreter
    # with debugging enabled (aka 'python-dbg' on debian).
    cxxflags += ['-g0']

elif platform.system()=='Darwin':
    # avoid later failure where install_name_tool may run out of space.
    #   install_name_tool: changing install names or rpaths can't be redone for:
    #   ... because larger updated load commands do not fit (the program must be relinked,
    #   and you may need to use -headerpad or -headerpad_max_install_names)
    ldflags += ['-Wl,-headerpad_max_install_names']

ext = Extension(
    name='p4p._p4p',
    sources = [
        "src/p4p_top.cpp",
        "src/p4p_type.cpp",
        "src/p4p_value.cpp",
        "src/p4p_array.cpp",

        "src/p4p_server.cpp",
        "src/p4p_server_provider.cpp",
        "src/p4p_server_sharedpv.cpp",

        "src/p4p_client.cpp",
    ],
    include_dirs = get_numpy_include_dirs()+[epicscorelibs.path.include_path],
    define_macros = get_config_var('CPPFLAGS'),
    extra_compile_args = get_config_var('CXXFLAGS')+cxxflags,
    extra_link_args = get_config_var('LDFLAGS')+ldflags,
    dsos = ['epicscorelibs.lib.pvAccess',
            'epicscorelibs.lib.pvData',
            'epicscorelibs.lib.ca',
            'epicscorelibs.lib.Com'
    ],
    libraries = get_config_var('LDADD'),
)

gwext = cythonize([
    Extension(
        name='p4p._gw',
        sources=['src/p4p/_gw.pyx', 'src/gwchannel.cpp'],
        include_dirs = get_numpy_include_dirs()+[epicscorelibs.path.include_path, 'src', 'src/p4p'],
        define_macros = get_config_var('CPPFLAGS'),
        extra_compile_args = get_config_var('CXXFLAGS')+cxxflags,
        extra_link_args = get_config_var('LDFLAGS')+ldflags,
        dsos = ['epicscorelibs.lib.pvAccess',
            'epicscorelibs.lib.pvData',
            'epicscorelibs.lib.ca',
            'epicscorelibs.lib.Com'
        ],
        libraries = get_config_var('LDADD'),
    )
])

setup(
    name='p4p',
    version=package_version,
    description="Python interface to PVAccess protocol client",
    url='https://mdavidsaver.github.io/p4p',
    author='Michael Davidsaver',
    author_email='mdavidsaver@gmail.com',
    license='BSD',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Distributed Computing',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
    ],
    keywords='epics scada',
    python_requires='>=2.7',

    packages=[
        'p4p',
        'p4p.nt',
        'p4p.client',
        'p4p.test',
        'p4p.server',
        'p4p.asLib',
    ],
    package_dir={'':'src'},
    ext_modules = [ext]+gwext,
    install_requires = [
        epicscorelibs.version.abi_requires(),
        # assume ABI forward compatibility as indicated by
        # https://github.com/numpy/numpy/blob/master/numpy/core/setup_common.py#L28
        'numpy >=%s'%numpy.version.short_version,
        'nose>=1.1.2',
        'ply', # for asLib
    ],
    extras_require={
        'qt': ['qtpy'],
    },
    entry_points = {
        'console_scripts': ['pvagw=p4p.gw:main'],
    },
    zip_safe = False,
)




import logging
import warnings
_log = logging.getLogger(__name__)

import unittest
import sys
import random
import weakref
import gc
from unittest.case import SkipTest

from ..nt import NTScalar, NTURI
from ..server import Server, StaticProvider
from .utils import RefTestCase

try:
    import cothread
except ImportError:
    raise SkipTest('No cothread')
else:
    from ..client.cothread import Context, Disconnected, RemoteError
    from ..server.cothread import SharedPV, _sync
    from ..server import cothread as srv_cothread

    class GPMHandler:
        def put(self, pv, op):
            _log.debug("putting %s <- %s", op.name(), op.value())
            cothread.Yield()  # because we can
            pv.post(op.value() * 2)
            op.done()

        def rpc(self, pv, op):
            V = op.value()
            if V.query.get('oops'):
                op.done(error='oops')
            else:
                op.done(NTScalar('i').wrap(42))

    class TestGPM(RefTestCase):
        timeout = 1.0

        def _sleep(self, delay):
            cothread.Sleep(delay)

        def setUp(self):
            super(TestGPM, self).setUp()

            self.pv = SharedPV(nt=NTScalar('i'), initial=0, handler=GPMHandler())
            self.pv2 = SharedPV(handler=GPMHandler(), nt=NTScalar('d'), initial=42.0)
            self.provider = StaticProvider("serverend")
            self.provider.add('foo', self.pv)
            self.provider.add('bar', self.pv2)

        def tearDown(self):
            del self.pv
            del self.pv2
            del self.provider
            _sync()
            gc.collect()
            self.assertSetEqual(set(srv_cothread._handlers), set()) # ensure no outstanding server handlers
            super(TestGPM, self).tearDown()

        def test_getput(self):
            with Server(providers=[self.provider], isolate=True) as S:
                with Context('pva', conf=S.conf(), useenv=False) as C:
                    self.assertEqual(0, C.get('foo'))

                    C.put('foo', 5)

                    self.assertEqual(5 * 2, C.get('foo'))

                    self.assertEqual([5 * 2, 42.0], C.get(['foo', 'bar']))

                    C.put(['foo', 'bar'], [6, 7])

                    self.assertEqual([6 * 2, 7 * 2], C.get(['foo', 'bar']))

        def test_monitor(self):
            with Server(providers=[self.provider], isolate=True) as S:
                with Context('pva', conf=S.conf(), useenv=False) as C:

                    Q = cothread.EventQueue()

                    with C.monitor('foo', Q.Signal, notify_disconnect=True) as sub:
                        self.assertIsInstance(Q.Wait(timeout=self.timeout), Disconnected)

                        self.assertEqual(0, Q.Wait(timeout=self.timeout))

                        C.put('foo', 2)

                        self.assertEqual(2 * 2, Q.Wait(timeout=self.timeout))

                        self.pv.close()

                        self.assertIsInstance(Q.Wait(timeout=self.timeout), Disconnected)

                        self.pv.open(3)

                        self.assertEqual(3, Q.Wait(timeout=self.timeout))

    class TestRPC(RefTestCase):
        timeout = 1.0

        def _sleep(self, delay):
            cothread.Sleep(delay)

        def setUp(self):
            super(TestRPC, self).setUp()

            self.pv = SharedPV(nt=NTScalar('i'), initial=0, handler=GPMHandler())
            self.provider = StaticProvider("serverend")
            self.provider.add('foo', self.pv)

        def tearDown(self):
            del self.pv
            del self.provider
            _sync()
            gc.collect()
            self.assertSetEqual(set(srv_cothread._handlers), set()) # ensure no outstanding server handlers
            super(TestRPC, self).tearDown()

        def test_rpc(self):
            with Server(providers=[self.provider], isolate=True) as S:
                with Context('pva', conf=S.conf(), useenv=False) as C:

                    args = NTURI([
                        ('lhs', 'd'),
                        ('rhs', 'd'),
                    ])

                    ret = C.rpc('foo', args.wrap('foo', kws={'lhs':1, 'rhs':2}))
                    _log.debug("RET %s", ret)
                    self.assertEqual(ret, 42)

        def test_rpc_error(self):
            with Server(providers=[self.provider], isolate=True) as S:
                with Context('pva', conf=S.conf(), useenv=False) as C:

                    args = NTURI([
                        ('oops', '?'),
                    ])

                    with self.assertRaisesRegexp(RemoteError, 'oops'):
                        ret = C.rpc('foo', args.wrap('foo', kws={'oops':True}))

    class TestFirstLast(RefTestCase):
        maxDiff = 1000
        timeout = 1.0
        mode = 'Mask'

        class Handler:
            def __init__(self):
                self.evt = cothread.Event(auto_reset=False)
                self.conn = None
            def onFirstConnect(self, pv):
                _log.debug("onFirstConnect")
                self.conn = True
                self.evt.Signal()
            def onLastDisconnect(self, pv):
                _log.debug("onLastDisconnect")
                self.conn = False
                self.evt.Signal()

        def _sleep(self, delay):
            cothread.Sleep(delay)

        def setUp(self):
            # gc.set_debug(gc.DEBUG_LEAK)
            super(TestFirstLast, self).setUp()

            self.H = self.Handler()
            self.pv = SharedPV(handler=self.H,
                            nt=NTScalar('d'),
                            options={'mapperMode':self.mode})
            self.sprov = StaticProvider("serverend")
            self.sprov.add('foo', self.pv)

            self.server = Server(providers=[self.sprov], isolate=True)

        def tearDown(self):
            self.server.stop()
            del self.server
            del self.sprov
            del self.pv
            del self.H
            _sync()
            gc.collect()
            self.assertSetEqual(set(srv_cothread._handlers), set()) # ensure no outstanding server handlers
            super(TestFirstLast, self).tearDown()

        def testClientDisconn(self):
            self.pv.open(1.0)

            with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
                Q = cothread.EventQueue()
                with ctxt.monitor('foo', Q.Signal, notify_disconnect=True):

                    Q.Wait(timeout=self.timeout) # initial update

                    _log.debug('TEST')
                    self.H.evt.Wait(self.timeout) # onFirstConnect()
                    self.H.evt.Reset()
                    self.assertTrue(self.H.conn)

            self.H.evt.Wait(self.timeout) # onLastDisconnect()
            _log.debug('SHUTDOWN')
            self.assertFalse(self.H.conn)

        def testServerShutdown(self):
            self.pv.open(1.0)

            with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
                Q = cothread.EventQueue()
                with ctxt.monitor('foo', Q.Signal, notify_disconnect=True):

                    Q.Wait(timeout=self.timeout)

                    _log.debug('TEST')
                    self.H.evt.Wait(self.timeout) # initial update
                    self.H.evt.Reset()
                    self.assertIs(self.H.conn, True)

                    self.server.stop()

                    self.H.evt.Wait(self.timeout) # wait for
                    _log.debug('SHUTDOWN')
                    self.assertIs(self.H.conn, False)

        def testPVClose(self):
            self.pv.open(1.0)

            with Context('pva', conf=self.server.conf(), useenv=False, unwrap={}) as ctxt:
                Q = cothread.EventQueue()
                with ctxt.monitor('foo', Q.Signal, notify_disconnect=True):

                    Q.Wait(timeout=self.timeout) # initial update

                    _log.debug('TEST')
                    self.H.evt.Wait(self.timeout)
                    self.H.evt.Reset()
                    self.assertTrue(self.H.conn)

                    _log.debug('CLOSING')
                    self.sprov.remove('foo') # prevent new connections while destroying
                    self.pv.close(destroy=True, sync=True, timeout=self.timeout)

                    _log.debug('CLOSED')
                    self.assertFalse(self.H.conn)
